const char net_box_lua[] =
"-- net_box.lua (internal file)\n"
"local log      = require('log')\n"
"local ffi      = require('ffi')\n"
"local buffer   = require('buffer')\n"
"local socket   = require('socket')\n"
"local fiber    = require('fiber')\n"
"local msgpack  = require('msgpack')\n"
"local errno    = require('errno')\n"
"local urilib   = require('uri')\n"
"local internal = require('net.box.lib')\n"
"local trigger  = require('internal.trigger')\n"
"\n"
"local band              = bit.band\n"
"local max               = math.max\n"
"local fiber_clock       = fiber.clock\n"
"local fiber_self        = fiber.self\n"
"local decode            = msgpack.decode_unchecked\n"
"local decode_map_header = msgpack.decode_map_header\n"
"local buffer_reg        = buffer.reg1\n"
"\n"
"local table_new           = require('table.new')\n"
"local check_iterator_type = box.internal.check_iterator_type\n"
"local check_index_arg     = box.internal.check_index_arg\n"
"local check_space_arg     = box.internal.check_space_arg\n"
"local check_primary_index = box.internal.check_primary_index\n"
"\n"
"local communicate     = internal.communicate\n"
"local encode_auth     = internal.encode_auth\n"
"local encode_select   = internal.encode_select\n"
"local decode_greeting = internal.decode_greeting\n"
"\n"
"local TIMEOUT_INFINITY = 500 * 365 * 86400\n"
"local VSPACE_ID        = 281\n"
"local VINDEX_ID        = 289\n"
"local VCOLLATION_ID    = 277\n"
"local DEFAULT_CONNECT_TIMEOUT = 10\n"
"\n"
"local IPROTO_STATUS_KEY    = 0x00\n"
"local IPROTO_ERRNO_MASK    = 0x7FFF\n"
"local IPROTO_SYNC_KEY      = 0x01\n"
"local IPROTO_SCHEMA_VERSION_KEY = 0x05\n"
"local IPROTO_METADATA_KEY = 0x32\n"
"local IPROTO_SQL_INFO_KEY = 0x42\n"
"local SQL_INFO_ROW_COUNT_KEY = 0\n"
"local IPROTO_FIELD_NAME_KEY = 0\n"
"local IPROTO_DATA_KEY      = 0x30\n"
"local IPROTO_ERROR_24      = 0x31\n"
"local IPROTO_ERROR         = 0x52\n"
"local IPROTO_GREETING_SIZE = 128\n"
"local IPROTO_CHUNK_KEY     = 128\n"
"local IPROTO_OK_KEY        = 0\n"
"\n"
"-- select errors from box.error\n"
"local E_UNKNOWN              = box.error.UNKNOWN\n"
"local E_NO_CONNECTION        = box.error.NO_CONNECTION\n"
"local E_TIMEOUT              = box.error.TIMEOUT\n"
"local E_PROC_LUA             = box.error.PROC_LUA\n"
"local E_NO_SUCH_SPACE        = box.error.NO_SUCH_SPACE\n"
"\n"
"ffi.cdef[[\n"
"struct error *\n"
"error_unpack_unsafe(const char **data);\n"
"\n"
"void\n"
"error_unref(struct error *e);\n"
"]]\n"
"\n"
"-- utility tables\n"
"local is_final_state         = {closed = 1, error = 1}\n"
"\n"
"local function decode_nil(raw_data, raw_data_end)\n"
"    return nil, raw_data_end\n"
"end\n"
"local function decode_data(raw_data)\n"
"    local response, raw_end = decode(raw_data)\n"
"    return response[IPROTO_DATA_KEY], raw_end\n"
"end\n"
"local function decode_tuple(raw_data, raw_data_end, format)\n"
"    local response, raw_end = internal.decode_select(raw_data, nil, format)\n"
"    return response[1], raw_end\n"
"end\n"
"local function decode_get(raw_data, raw_data_end, format)\n"
"    local body, raw_end = internal.decode_select(raw_data, nil, format)\n"
"    if body[2] then\n"
"        return nil, raw_end, box.error.MORE_THAN_ONE_TUPLE\n"
"    end\n"
"    return body[1], raw_end\n"
"end\n"
"local function decode_count(raw_data)\n"
"    local response, raw_end = decode(raw_data)\n"
"    return response[IPROTO_DATA_KEY][1], raw_end\n"
"end\n"
"local function decode_push(raw_data)\n"
"    local response, raw_end = decode(raw_data)\n"
"    return response[IPROTO_DATA_KEY][1], raw_end\n"
"end\n"
"\n"
"local function version_id(major, minor, patch)\n"
"    return bit.bor(bit.lshift(major, 16), bit.lshift(minor, 8), patch)\n"
"end\n"
"\n"
"local function version_at_least(peer_version_id, major, minor, patch)\n"
"    return peer_version_id >= version_id(major, minor, patch)\n"
"end\n"
"\n"
"local method_encoder = {\n"
"    ping    = internal.encode_ping,\n"
"    call_16 = internal.encode_call_16,\n"
"    call_17 = internal.encode_call,\n"
"    eval    = internal.encode_eval,\n"
"    insert  = internal.encode_insert,\n"
"    replace = internal.encode_replace,\n"
"    delete  = internal.encode_delete,\n"
"    update  = internal.encode_update,\n"
"    upsert  = internal.encode_upsert,\n"
"    select  = internal.encode_select,\n"
"    execute = internal.encode_execute,\n"
"    prepare = internal.encode_prepare,\n"
"    unprepare = internal.encode_prepare,\n"
"    get     = internal.encode_select,\n"
"    min     = internal.encode_select,\n"
"    max     = internal.encode_select,\n"
"    count   = internal.encode_call,\n"
"    -- inject raw data into connection, used by console and tests\n"
"    inject = function(buf, id, bytes)\n"
"        local ptr = buf:reserve(#bytes)\n"
"        ffi.copy(ptr, bytes, #bytes)\n"
"        buf.wpos = ptr + #bytes\n"
"    end\n"
"}\n"
"\n"
"local method_decoder = {\n"
"    ping    = decode_nil,\n"
"    call_16 = internal.decode_select,\n"
"    call_17 = decode_data,\n"
"    eval    = decode_data,\n"
"    insert  = decode_tuple,\n"
"    replace = decode_tuple,\n"
"    delete  = decode_tuple,\n"
"    update  = decode_tuple,\n"
"    upsert  = decode_nil,\n"
"    select  = internal.decode_select,\n"
"    execute = internal.decode_execute,\n"
"    prepare = internal.decode_prepare,\n"
"    unprepare = decode_nil,\n"
"    get     = decode_get,\n"
"    min     = decode_get,\n"
"    max     = decode_get,\n"
"    count   = decode_count,\n"
"    inject  = decode_data,\n"
"    push    = decode_push,\n"
"}\n"
"\n"
"local function decode_error(raw_data)\n"
"    local ptr = buffer_reg.acucp\n"
"    ptr[0] = raw_data\n"
"    local err = ffi.C.error_unpack_unsafe(ptr)\n"
"    if err ~= nil then\n"
"        err._refs = err._refs + 1\n"
"        err = ffi.gc(err, ffi.C.error_unref)\n"
"        -- From FFI it is returned as 'struct error *', which is\n"
"        -- not considered equal to 'const struct error &', and is\n"
"        -- is not accepted by functions like box.error(). Need to\n"
"        -- cast explicitly.\n"
"        err = ffi.cast('const struct error &', err)\n"
"    else\n"
"        -- Error unpacker installs fail reason into diag.\n"
"        box.error()\n"
"    end\n"
"    return err, ptr[0]\n"
"end\n"
"\n"
"local response_decoder = {\n"
"    [IPROTO_ERROR_24] = decode,\n"
"    [IPROTO_ERROR] = decode_error,\n"
"}\n"
"\n"
"local function next_id(id) return band(id + 1, 0x7FFFFFFF) end\n"
"\n"
"--\n"
"-- Connect to a remote server, do handshake.\n"
"-- @param host Hostname.\n"
"-- @param port TCP port.\n"
"-- @param timeout Timeout to connect and receive greeting.\n"
"--\n"
"-- @retval nil, err Error occured. The reason is returned.\n"
"-- @retval two non-nils A connected socket and a decoded greeting.\n"
"--\n"
"local function establish_connection(host, port, timeout)\n"
"    local timeout = timeout or DEFAULT_CONNECT_TIMEOUT\n"
"    local begin = fiber.clock()\n"
"    local s = socket.tcp_connect(host, port, timeout)\n"
"    if not s then\n"
"        return nil, errno.strerror(errno())\n"
"    end\n"
"    local msg = s:read({chunk = IPROTO_GREETING_SIZE},\n"
"                        timeout - (fiber.clock() - begin))\n"
"    if not msg then\n"
"        local err = s:error()\n"
"        s:close()\n"
"        return nil, err\n"
"    end\n"
"    local greeting, err = decode_greeting(msg)\n"
"    if not greeting then\n"
"        s:close()\n"
"        return nil, err\n"
"    end\n"
"    return s, greeting\n"
"end\n"
"\n"
"--\n"
"-- Default action on push during a synchronous request -\n"
"-- ignore.\n"
"--\n"
"local function on_push_sync_default(...) end\n"
"\n"
"--\n"
"-- Basically, *transport* is a TCP connection speaking one of\n"
"-- Tarantool network protocols. This is a low-level interface.\n"
"-- Primary features:\n"
"--  * implements protocols; concurrent perform_request()-s benefit from\n"
"--    multiplexing support in the protocol;\n"
"--  * schema-aware (optional) - snoops responses and initiates\n"
"--    schema reload when a response has a new schema version;\n"
"--  * delivers transport events via the callback.\n"
"--\n"
"-- Transport state machine:\n"
"--\n"
"-- State machine starts in 'initial' state. New_sm method\n"
"-- accepts an established connection and spawns a worker fiber.\n"
"-- Stop method sets the state to 'closed' and kills the worker.\n"
"-- If the transport is already in 'error' state stop() does\n"
"-- nothing.\n"
"--\n"
"-- State chart:\n"
"--\n"
"-- connecting -> initial +-> active\n"
"--                        \\\n"
"--                         +-> auth -> fetch_schema <-> active\n"
"--\n"
"--  (any state, on error) -> error_reconnect -> connecting -> ...\n"
"--                                           \\\n"
"--                                             -> [error]\n"
"--  (any_state, but [error]) -> [closed]\n"
"--\n"
"--\n"
"-- State change events can be delivered to the transport user via\n"
"-- the optional 'callback' argument:\n"
"--\n"
"-- The callback functions needs to have the following signature:\n"
"--\n"
"--  callback(event_name, ...)\n"
"--\n"
"-- The following events are delivered, with arguments:\n"
"--\n"
"--  'state_changed', state, errno, error\n"
"--  'handshake', greeting -> nil (accept) / errno, error (reject)\n"
"--  'will_fetch_schema'   -> true (approve) / false (skip fetch)\n"
"--  'did_fetch_schema', schema_version, spaces, indices\n"
"--  'reconnect_timeout'   -> get reconnect timeout if set and > 0,\n"
"--                           else nil is returned.\n"
"--\n"
"-- Suggestion for callback writers: sleep a few secs before approving\n"
"-- reconnect.\n"
"--\n"
"local function create_transport(host, port, user, password, callback,\n"
"                                connection, greeting)\n"
"    -- check / normalize credentials\n"
"    if user == nil and password ~= nil then\n"
"        box.error(E_PROC_LUA, 'net.box: user is not defined')\n"
"    end\n"
"    if user ~= nil and password == nil then password = '' end\n"
"\n"
"    -- Current state machine's state.\n"
"    local state            = 'initial'\n"
"    local last_errno\n"
"    local last_error\n"
"    local state_cond       = fiber.cond() -- signaled when the state changes\n"
"\n"
"    -- Async requests currently 'in flight', keyed by a request\n"
"    -- id. Value refs are weak hence if a client dies\n"
"    -- unexpectedly, GC cleans the mess.\n"
"    -- Async request can not be timed out completely. Instead a\n"
"    -- user must decide when he does not want to wait for\n"
"    -- response anymore.\n"
"    -- Sync requests are implemented as async call + immediate\n"
"    -- wait for a result.\n"
"    local requests         = setmetatable({}, { __mode = 'v' })\n"
"    local next_request_id  = 1\n"
"\n"
"    local worker_fiber\n"
"    local send_buf         = buffer.ibuf(buffer.READAHEAD)\n"
"    local recv_buf         = buffer.ibuf(buffer.READAHEAD)\n"
"\n"
"    --\n"
"    -- Async request metamethods.\n"
"    --\n"
"    local request_index = {}\n"
"    --\n"
"    -- When an async request is finalized (with ok or error - no\n"
"    -- matter), its 'id' field is nullified by a response\n"
"    -- dispatcher.\n"
"    --\n"
"    function request_index:is_ready()\n"
"        return self.id == nil or worker_fiber == nil\n"
"    end\n"
"    --\n"
"    -- When a request is finished, a result can be got from a\n"
"    -- future object anytime.\n"
"    -- @retval result, nil Success, the response is returned.\n"
"    -- @retval nil, error Error occured.\n"
"    --\n"
"    function request_index:result()\n"
"        if self.errno then\n"
"            if type(self.response) ~= 'cdata' then\n"
"                -- Error could be set by the connection state\n"
"                -- machine, and it is usually a string explaining\n"
"                -- a reason.\n"
"                self.response = box.error.new({code = self.errno,\n"
"                                               reason = self.response})\n"
"            end\n"
"            return nil, self.response\n"
"        elseif not self.id then\n"
"            return self.response\n"
"        elseif not worker_fiber then\n"
"            return nil, box.error.new(E_NO_CONNECTION)\n"
"        else\n"
"            return nil, box.error.new(box.error.PROC_LUA,\n"
"                                      'Response is not ready')\n"
"        end\n"
"    end\n"
"    --\n"
"    -- Get the next message or the final result.\n"
"    -- @param iterator Iterator object.\n"
"    -- @param i Index to get a next message from.\n"
"    --\n"
"    -- @retval nil, nil The request is finished.\n"
"    -- @retval i + 1, object A message/response and its index.\n"
"    -- @retval box.NULL, error An error occured. When this\n"
"    --         function is called in 'for k, v in future:pairs()',\n"
"    --         `k` becomes box.NULL, and `v` becomes error object.\n"
"    --         On error the key becomes exactly box.NULL instead\n"
"    --         of nil, because nil is treated by Lua as iteration\n"
"    --         end marker. Nil does not participate in iteration,\n"
"    --         and does not allow to continue it.\n"
"    --\n"
"    local function request_iterator_next(iterator, i)\n"
"        if i == box.NULL then\n"
"            return nil, nil\n"
"        else\n"
"            i = i + 1\n"
"        end\n"
"        local request = iterator.request\n"
"        local messages = request.on_push_ctx\n"
"    ::retry::\n"
"        if i <= #messages then\n"
"            return i, messages[i]\n"
"        end\n"
"        if request:is_ready() then\n"
"            -- After all the messages are iterated, `i` is equal\n"
"            -- to #messages + 1. After response reading `i`\n"
"            -- becomes #messages + 2. It is the trigger to finish\n"
"            -- the iteration.\n"
"            if i > #messages + 1 then\n"
"                return nil, nil\n"
"            end\n"
"            local response, err = request:result()\n"
"            if err then\n"
"                return box.NULL, err\n"
"            end\n"
"            return i, response\n"
"        end\n"
"        local old_message_count = #messages\n"
"        local timeout = iterator.timeout\n"
"        repeat\n"
"            local ts = fiber_clock()\n"
"            request.cond:wait(timeout)\n"
"            timeout = timeout - (fiber_clock() - ts)\n"
"            if request:is_ready() or old_message_count ~= #messages then\n"
"                goto retry\n"
"            end\n"
"        until timeout <= 0\n"
"        return box.NULL, box.error.new(E_TIMEOUT)\n"
"    end\n"
"    --\n"
"    -- Iterate over all messages, received by a request. @Sa\n"
"    -- request_iterator_next for details what to expect in `for`\n"
"    -- key/value pairs.\n"
"    -- @param timeout One iteration timeout.\n"
"    -- @retval next() callback, iterator, zero key.\n"
"    --\n"
"    function request_index:pairs(timeout)\n"
"        if timeout then\n"
"            if type(timeout) ~= 'number' or timeout < 0 then\n"
"                error('Usage: future:pairs(timeout)')\n"
"            end\n"
"        else\n"
"            timeout = TIMEOUT_INFINITY\n"
"        end\n"
"        local iterator = {request = self, timeout = timeout}\n"
"        return request_iterator_next, iterator, 0\n"
"    end\n"
"    --\n"
"    -- Wait for a response or error max timeout seconds.\n"
"    -- @param timeout Max seconds to wait.\n"
"    -- @retval result, nil Success, the response is returned.\n"
"    -- @retval nil, error Error occured.\n"
"    --\n"
"    function request_index:wait_result(timeout)\n"
"        if timeout then\n"
"            if type(timeout) ~= 'number' or timeout < 0 then\n"
"                error('Usage: future:wait_result(timeout)')\n"
"            end\n"
"        else\n"
"            timeout = TIMEOUT_INFINITY\n"
"        end\n"
"        if not self:is_ready() then\n"
"            -- When a response is ready before timeout, the\n"
"            -- waiting client is waked up prematurely.\n"
"            while timeout > 0 and not self:is_ready() do\n"
"                local ts = fiber.clock()\n"
"                self.cond:wait(timeout)\n"
"                timeout = timeout - (fiber.clock() - ts)\n"
"            end\n"
"            if not self:is_ready() then\n"
"                return nil, box.error.new(E_TIMEOUT)\n"
"            end\n"
"        end\n"
"        return self:result()\n"
"    end\n"
"    --\n"
"    -- Make a connection forget about the response. When it will\n"
"    -- be received, it will be ignored. It reduces size of\n"
"    -- requests table speeding up other requests.\n"
"    --\n"
"    function request_index:discard()\n"
"        if self.id then\n"
"            requests[self.id] = nil\n"
"            self.id = nil\n"
"            self.errno = box.error.PROC_LUA\n"
"            self.response = 'Response is discarded'\n"
"        end\n"
"    end\n"
"\n"
"    local request_mt = { __index = request_index }\n"
"\n"
"    -- STATE SWITCHING --\n"
"    local function set_state(new_state, new_errno, new_error)\n"
"        state = new_state\n"
"        last_errno = new_errno\n"
"        last_error = new_error\n"
"        callback('state_changed', new_state, new_errno, new_error)\n"
"        state_cond:broadcast()\n"
"        if state == 'error' or state == 'error_reconnect' or\n"
"           state == 'closed' then\n"
"            for _, request in pairs(requests) do\n"
"                request.id = nil\n"
"                request.errno = new_errno\n"
"                request.response = new_error\n"
"                request.cond:broadcast()\n"
"            end\n"
"            requests = {}\n"
"        end\n"
"    end\n"
"\n"
"    -- FYI: [] on a string is valid\n"
"    local function wait_state(target_state, timeout)\n"
"        local deadline = fiber_clock() + (timeout or TIMEOUT_INFINITY)\n"
"        repeat until state == target_state or target_state[state] or\n"
"                     is_final_state[state] or\n"
"                     not state_cond:wait(max(0, deadline - fiber_clock()))\n"
"        return state == target_state or target_state[state] or false\n"
"    end\n"
"\n"
"    -- START/STOP --\n"
"    local protocol_sm\n"
"\n"
"    local function start()\n"
"        if state ~= 'initial' then return not is_final_state[state] end\n"
"        fiber.create(function()\n"
"            local ok, err, timeout\n"
"            worker_fiber = fiber_self()\n"
"            fiber.name(string.format('%s:%s (net.box)', host, port), {truncate=true})\n"
"            if not connection then\n"
"                goto do_connect\n"
"            end\n"
"    ::handle_connection::\n"
"            ok, err = pcall(protocol_sm)\n"
"            if not (ok or is_final_state[state]) then\n"
"                set_state('error', E_UNKNOWN, err)\n"
"            end\n"
"            if connection then\n"
"                connection:close()\n"
"                connection = nil\n"
"            end\n"
"            timeout = callback('reconnect_timeout')\n"
"    ::do_reconnect::\n"
"            if not timeout or state ~= 'error_reconnect' then\n"
"                goto stop\n"
"            end\n"
"            fiber.sleep(timeout)\n"
"            timeout = callback('reconnect_timeout')\n"
"            if not timeout or state ~= 'error_reconnect' then\n"
"                goto stop\n"
"            end\n"
"    ::do_connect::\n"
"            connection, greeting =\n"
"                establish_connection(host, port, callback('fetch_connect_timeout'))\n"
"            if connection then\n"
"                goto handle_connection\n"
"            end\n"
"            timeout = callback('reconnect_timeout')\n"
"            if not timeout then\n"
"                set_state('error', E_NO_CONNECTION, greeting)\n"
"                goto stop\n"
"            end\n"
"            set_state('error_reconnect', E_NO_CONNECTION, greeting)\n"
"            goto do_reconnect\n"
"    ::stop::\n"
"            send_buf:recycle()\n"
"            recv_buf:recycle()\n"
"            worker_fiber = nil\n"
"        end)\n"
"    end\n"
"\n"
"    local function stop()\n"
"        if not is_final_state[state] then\n"
"            set_state('closed', E_NO_CONNECTION, 'Connection closed')\n"
"        end\n"
"        if worker_fiber then\n"
"            worker_fiber:cancel()\n"
"            worker_fiber = nil\n"
"        end\n"
"    end\n"
"\n"
"    --\n"
"    -- Send a request and do not wait for response.\n"
"    -- @retval nil, error Error occured.\n"
"    -- @retval not nil Future object.\n"
"    --\n"
"    local function perform_async_request(buffer, skip_header, method, on_push,\n"
"                                         on_push_ctx, request_ctx, ...)\n"
"        if state ~= 'active' and state ~= 'fetch_schema' then\n"
"            return nil, box.error.new({code = last_errno or E_NO_CONNECTION,\n"
"                                       reason = last_error})\n"
"        end\n"
"        -- alert worker to notify it of the queued outgoing data;\n"
"        -- if the buffer wasn't empty, assume the worker was already alerted\n"
"        if send_buf:size() == 0 then\n"
"            worker_fiber:wakeup()\n"
"        end\n"
"        local id = next_request_id\n"
"        method_encoder[method](send_buf, id, ...)\n"
"        next_request_id = next_id(id)\n"
"        -- Request in most cases has maximum 10 members:\n"
"        -- method, buffer, skip_header, id, cond, errno, response,\n"
"        -- on_push, on_push_ctx and ctx.\n"
"        local request = setmetatable(table_new(0, 10), request_mt)\n"
"        request.method = method\n"
"        request.buffer = buffer\n"
"        request.skip_header = skip_header\n"
"        request.id = id\n"
"        request.cond = fiber.cond()\n"
"        requests[id] = request\n"
"        request.on_push = on_push\n"
"        request.on_push_ctx = on_push_ctx\n"
"        request.ctx = request_ctx\n"
"        return request\n"
"    end\n"
"\n"
"    --\n"
"    -- Send a request and wait for response.\n"
"    -- @retval nil, error Error occured.\n"
"    -- @retval not nil Response object.\n"
"    --\n"
"    local function perform_request(timeout, buffer, skip_header, method,\n"
"                                   on_push, on_push_ctx, request_ctx, ...)\n"
"        local request, err =\n"
"            perform_async_request(buffer, skip_header, method, on_push,\n"
"                                  on_push_ctx, request_ctx, ...)\n"
"        if not request then\n"
"            return nil, err\n"
"        end\n"
"        return request:wait_result(timeout)\n"
"    end\n"
"\n"
"    local function dispatch_response_iproto(hdr, body_rpos, body_end)\n"
"        local id = hdr[IPROTO_SYNC_KEY]\n"
"        local request = requests[id]\n"
"        if request == nil then -- nobody is waiting for the response\n"
"            return\n"
"        end\n"
"        local status = hdr[IPROTO_STATUS_KEY]\n"
"        local body\n"
"        local body_len = body_end - body_rpos\n"
"\n"
"        if status > IPROTO_CHUNK_KEY then\n"
"            -- Handle errors\n"
"            requests[id] = nil\n"
"            request.id = nil\n"
"            local map_len, key, skip\n"
"            map_len, body_rpos = decode_map_header(body_rpos, body_len)\n"
"            -- Reserve for 2 keys and 2 array indexes. Because no\n"
"            -- any guarantees how Lua will decide to save the\n"
"            -- sparse table.\n"
"            body = table_new(2, 2)\n"
"            for _ = 1, map_len do\n"
"                key, body_rpos = decode(body_rpos)\n"
"                local rdec = response_decoder[key]\n"
"                if rdec then\n"
"                    body[key], body_rpos = rdec(body_rpos)\n"
"                else\n"
"                    skip, body_rpos = decode(body_rpos)\n"
"                end\n"
"            end\n"
"            assert(body_end == body_rpos, \"invalid xrow length\")\n"
"            request.errno = band(status, IPROTO_ERRNO_MASK)\n"
"            -- IPROTO_ERROR comprises error encoded with\n"
"            -- IPROTO_ERROR_24, so we may ignore content of that\n"
"            -- key.\n"
"            if body[IPROTO_ERROR] ~= nil then\n"
"                request.response = body[IPROTO_ERROR]\n"
"                assert(type(request.response) == 'cdata')\n"
"            else\n"
"                request.response = box.error.new({\n"
"                    code = request.errno,\n"
"                    reason = body[IPROTO_ERROR_24]\n"
"                })\n"
"            end\n"
"            request.cond:broadcast()\n"
"            return\n"
"        end\n"
"\n"
"        local buffer = request.buffer\n"
"        if buffer ~= nil then\n"
"            -- Copy xrow.body to user-provided buffer\n"
"            if request.skip_header then\n"
"                -- Skip {[IPROTO_DATA_KEY] = ...} wrapper.\n"
"                local map_len, key\n"
"                map_len, body_rpos = decode_map_header(body_rpos, body_len)\n"
"                assert(map_len == 1)\n"
"                key, body_rpos = decode(body_rpos)\n"
"                assert(key == IPROTO_DATA_KEY)\n"
"                body_len = body_end - body_rpos\n"
"            end\n"
"            local wpos = buffer:alloc(body_len)\n"
"            ffi.copy(wpos, body_rpos, body_len)\n"
"            body_len = tonumber(body_len)\n"
"            if status == IPROTO_OK_KEY then\n"
"                request.response = body_len\n"
"                requests[id] = nil\n"
"                request.id = nil\n"
"            else\n"
"                request.on_push(request.on_push_ctx, body_len)\n"
"            end\n"
"            request.cond:broadcast()\n"
"            return\n"
"        end\n"
"\n"
"        local real_end\n"
"        -- Decode xrow.body[DATA] to Lua objects\n"
"        if status == IPROTO_OK_KEY then\n"
"            request.response, real_end, request.errno =\n"
"                method_decoder[request.method](body_rpos, body_end, request.ctx)\n"
"            assert(real_end == body_end, \"invalid body length\")\n"
"            requests[id] = nil\n"
"            request.id = nil\n"
"        else\n"
"            local msg\n"
"            msg, real_end, request.errno =\n"
"                method_decoder.push(body_rpos, body_end)\n"
"            assert(real_end == body_end, \"invalid body length\")\n"
"            request.on_push(request.on_push_ctx, msg)\n"
"        end\n"
"        request.cond:broadcast()\n"
"    end\n"
"\n"
"    local function new_request_id()\n"
"        local id = next_request_id;\n"
"        next_request_id = next_id(id)\n"
"        return id\n"
"    end\n"
"\n"
"    -- IO (WORKER FIBER) --\n"
"    local function send_and_recv(limit_or_boundary, timeout)\n"
"        return communicate(connection:fd(), send_buf, recv_buf,\n"
"                           limit_or_boundary, timeout)\n"
"    end\n"
"\n"
"    local function send_and_recv_iproto(timeout)\n"
"        local data_len = recv_buf.wpos - recv_buf.rpos\n"
"        local required = 0\n"
"        if data_len < 5 then\n"
"            required = 5\n"
"        else\n"
"            -- PWN! insufficient input validation\n"
"            local bufpos = recv_buf.rpos\n"
"            local len, rpos = decode(bufpos)\n"
"            required = (rpos - bufpos) + len\n"
"            if data_len >= required then\n"
"                local body_end = rpos + len\n"
"                local hdr, body_rpos = decode(rpos)\n"
"                recv_buf.rpos = body_end\n"
"                return nil, hdr, body_rpos, body_end\n"
"            end\n"
"        end\n"
"        local deadline = fiber_clock() + (timeout or TIMEOUT_INFINITY)\n"
"        local err, extra = send_and_recv(required, timeout)\n"
"        if err then\n"
"            return err, extra\n"
"        end\n"
"        return send_and_recv_iproto(max(0, deadline - fiber_clock()))\n"
"    end\n"
"\n"
"    local function send_and_recv_console(timeout)\n"
"        local delim = '\\n...\\n'\n"
"        local err, delim_pos = send_and_recv(delim, timeout)\n"
"        if err then\n"
"            return err, delim_pos\n"
"        else\n"
"            local response = ffi.string(recv_buf.rpos, delim_pos + #delim)\n"
"            recv_buf.rpos = recv_buf.rpos + delim_pos + #delim\n"
"            return nil, response\n"
"        end\n"
"    end\n"
"\n"
"    -- PROTOCOL STATE MACHINE (WORKER FIBER) --\n"
"    --\n"
"    -- The sm is implemented as a collection of functions performing\n"
"    -- tail-recursive calls to each other. Yep, Lua optimizes\n"
"    -- such calls, and yep, this is the canonical way to implement\n"
"    -- a state machine in Lua.\n"
"    local console_sm, iproto_auth_sm, iproto_schema_sm, iproto_sm, error_sm\n"
"\n"
"    --\n"
"    -- Protocol_sm is a core function of netbox. It calls all\n"
"    -- other ..._sm() functions, and explicitly or implicitly\n"
"    -- holds Lua referece on a connection object. It means, that\n"
"    -- until it works, the connection can not be garbage\n"
"    -- collected. See gh-3164, where because of reconnect sleeps\n"
"    -- in this function, a connection could not be deleted.\n"
"    --\n"
"    protocol_sm = function()\n"
"        assert(connection)\n"
"        assert(greeting)\n"
"        local err, msg = callback('handshake', greeting)\n"
"        if err then\n"
"            return error_sm(err, msg)\n"
"        end\n"
"        -- @deprecated since 1.10\n"
"        if greeting.protocol == 'Lua console' then\n"
"            log.warn(\"Netbox text protocol support is deprecated since 1.10, \"..\n"
"                     \"please use require('console').connect() instead\")\n"
"            local setup_delimiter = 'require(\"console\").delimiter(\"$EOF$\")\\n'\n"
"            method_encoder.inject(send_buf, nil, setup_delimiter)\n"
"            local err, response = send_and_recv_console()\n"
"            if err then\n"
"                return error_sm(err, response)\n"
"            elseif response ~= '---\\n...\\n' then\n"
"                return error_sm(E_NO_CONNECTION, 'Unexpected response')\n"
"            end\n"
"            local rid = next_request_id\n"
"            set_state('active')\n"
"            return console_sm(rid)\n"
"        elseif greeting.protocol == 'Binary' then\n"
"            return iproto_auth_sm(greeting.salt)\n"
"        else\n"
"            return error_sm(E_NO_CONNECTION,\n"
"                            'Unknown protocol: '..greeting.protocol)\n"
"        end\n"
"    end\n"
"\n"
"    console_sm = function(rid)\n"
"        local delim = '\\n...\\n'\n"
"        local err, response = send_and_recv_console()\n"
"        if err then\n"
"            return error_sm(err, response)\n"
"        else\n"
"            local request = requests[rid]\n"
"            if request == nil then -- nobody is waiting for the response\n"
"                return\n"
"            end\n"
"            request.id = nil\n"
"            requests[rid] = nil\n"
"            request.response = response\n"
"            request.cond:broadcast()\n"
"            return console_sm(next_id(rid))\n"
"        end\n"
"    end\n"
"\n"
"    iproto_auth_sm = function(salt)\n"
"        set_state('auth')\n"
"        if not user or not password then\n"
"            set_state('fetch_schema')\n"
"            return iproto_schema_sm()\n"
"        end\n"
"        encode_auth(send_buf, new_request_id(), user, password, salt)\n"
"        local err, hdr, body_rpos, body_end = send_and_recv_iproto()\n"
"        if err then\n"
"            return error_sm(err, hdr)\n"
"        end\n"
"        if hdr[IPROTO_STATUS_KEY] ~= 0 then\n"
"            local body\n"
"            body, body_end = decode(body_rpos)\n"
"            return error_sm(E_NO_CONNECTION, body[IPROTO_ERROR_24])\n"
"        end\n"
"        set_state('fetch_schema')\n"
"        return iproto_schema_sm(hdr[IPROTO_SCHEMA_VERSION_KEY])\n"
"    end\n"
"\n"
"    iproto_schema_sm = function(schema_version)\n"
"        if not callback('will_fetch_schema') then\n"
"            set_state('active')\n"
"            return iproto_sm(schema_version)\n"
"        end\n"
"        -- _vcollation view was added in 2.2.0-389-g3e3ef182f\n"
"        local peer_has_vcollation = version_at_least(greeting.version_id,\n"
"                                                     2, 2, 1)\n"
"        local select1_id = new_request_id()\n"
"        local select2_id = new_request_id()\n"
"        local select3_id\n"
"        local response = {}\n"
"        -- fetch everything from space _vspace, 2 = ITER_ALL\n"
"        encode_select(send_buf, select1_id, VSPACE_ID, 0, 2, 0, 0xFFFFFFFF, nil)\n"
"        -- fetch everything from space _vindex, 2 = ITER_ALL\n"
"        encode_select(send_buf, select2_id, VINDEX_ID, 0, 2, 0, 0xFFFFFFFF, nil)\n"
"        -- fetch everything from space _vcollation, 2 = ITER_ALL\n"
"        if peer_has_vcollation then\n"
"            select3_id = new_request_id()\n"
"            encode_select(send_buf, select3_id, VCOLLATION_ID, 0, 2, 0,\n"
"                          0xFFFFFFFF, nil)\n"
"        end\n"
"\n"
"        schema_version = nil -- any schema_version will do provided that\n"
"                             -- it is consistent across responses\n"
"        repeat\n"
"            local err, hdr, body_rpos, body_end = send_and_recv_iproto()\n"
"            if err then return error_sm(err, hdr) end\n"
"            dispatch_response_iproto(hdr, body_rpos, body_end)\n"
"            local id = hdr[IPROTO_SYNC_KEY]\n"
"            -- trick: omit check for peer_has_vcollation: id is\n"
"            -- not nil\n"
"            if id == select1_id or id == select2_id or id == select3_id then\n"
"                -- response to a schema query we've submitted\n"
"                local status = hdr[IPROTO_STATUS_KEY]\n"
"                local response_schema_version = hdr[IPROTO_SCHEMA_VERSION_KEY]\n"
"                if status ~= 0 then\n"
"                    -- No _vcollation space (server has an old\n"
"                    -- schema version).\n"
"                    local errno = band(status, IPROTO_ERRNO_MASK)\n"
"                    if id == select3_id and errno == E_NO_SUCH_SPACE then\n"
"                        peer_has_vcollation = false\n"
"                        goto continue\n"
"                    end\n"
"                    local body\n"
"                    body, body_end = decode(body_rpos)\n"
"                    return error_sm(E_NO_CONNECTION, body[IPROTO_ERROR_24])\n"
"                end\n"
"                if schema_version == nil then\n"
"                    schema_version = response_schema_version\n"
"                elseif schema_version ~= response_schema_version then\n"
"                    -- schema changed while fetching schema; restart loader\n"
"                    return iproto_schema_sm()\n"
"                end\n"
"                local body\n"
"                body, body_end = decode(body_rpos)\n"
"                response[id] = body[IPROTO_DATA_KEY]\n"
"            end\n"
"            ::continue::\n"
"        until response[select1_id] and response[select2_id] and\n"
"              (not peer_has_vcollation or response[select3_id])\n"
"        -- trick: response[select3_id] is nil when the key is nil\n"
"        callback('did_fetch_schema', schema_version, response[select1_id],\n"
"                 response[select2_id], response[select3_id])\n"
"        set_state('active')\n"
"        return iproto_sm(schema_version)\n"
"    end\n"
"\n"
"    iproto_sm = function(schema_version)\n"
"        local err, hdr, body_rpos, body_end = send_and_recv_iproto()\n"
"        if err then return error_sm(err, hdr) end\n"
"        dispatch_response_iproto(hdr, body_rpos, body_end)\n"
"        local status = hdr[IPROTO_STATUS_KEY]\n"
"        local response_schema_version = hdr[IPROTO_SCHEMA_VERSION_KEY]\n"
"        if response_schema_version > 0 and\n"
"           response_schema_version ~= schema_version then\n"
"            -- schema_version has been changed - start to load a new version.\n"
"            -- Sic: self.schema_version will be updated only after reload.\n"
"            local body\n"
"            body, body_end = decode(body_rpos)\n"
"            set_state('fetch_schema')\n"
"            return iproto_schema_sm(schema_version)\n"
"        end\n"
"        return iproto_sm(schema_version)\n"
"    end\n"
"\n"
"    error_sm = function(err, msg)\n"
"        if connection then connection:close(); connection = nil end\n"
"        send_buf:recycle()\n"
"        recv_buf:recycle()\n"
"        if state ~= 'closed' then\n"
"            if callback('reconnect_timeout') then\n"
"                set_state('error_reconnect', err, msg)\n"
"            else\n"
"                set_state('error', err, msg)\n"
"            end\n"
"        end\n"
"    end\n"
"\n"
"    return {\n"
"        stop            = stop,\n"
"        start           = start,\n"
"        wait_state      = wait_state,\n"
"        perform_request = perform_request,\n"
"        perform_async_request = perform_async_request,\n"
"    }\n"
"end\n"
"\n"
"-- Wrap create_transport, adding auto-stop-on-GC feature.\n"
"-- All the GC magic is neatly encapsulated!\n"
"-- The tricky part is the callback:\n"
"--  * callback (typically) references the transport (indirectly);\n"
"--  * worker fiber references the callback;\n"
"--  * fibers are GC roots - i.e. transport is never GC-ed!\n"
"-- We solve the issue by making the worker->callback ref weak.\n"
"-- Now it is necessary to have a strong ref to callback somewhere or\n"
"-- it is GC-ed prematurely. We wrap stop() method, stashing the\n"
"-- ref in an upvalue (stop() performance doesn't matter much.)\n"
"local create_transport = function(host, port, user, password, callback,\n"
"                                  connection, greeting)\n"
"    local weak_refs = setmetatable({callback = callback}, {__mode = 'v'})\n"
"    local function weak_callback(...)\n"
"        local callback = weak_refs.callback\n"
"        if callback then return callback(...) end\n"
"    end\n"
"    local transport = create_transport(host, port, user, password,\n"
"                                       weak_callback, connection, greeting)\n"
"    local transport_stop = transport.stop\n"
"    local gc_hook = ffi.gc(ffi.new('char[1]'), function()\n"
"        pcall(transport_stop)\n"
"    end)\n"
"    transport.stop = function()\n"
"        -- dummy gc_hook, callback refs prevent premature GC\n"
"        return transport_stop(gc_hook, callback)\n"
"    end\n"
"    return transport\n"
"end\n"
"\n"
"local function parse_connect_params(host_or_uri, ...) -- self\? host_or_uri port\? opts\?\n"
"    local port, opts = ...\n"
"    if type(host_or_uri) == 'table' then host_or_uri, port, opts = ... end\n"
"    if type(port) == 'table' then opts = port; port = nil end\n"
"    if opts == nil then opts = {} else\n"
"        local copy = {}\n"
"        for k, v in pairs(opts) do copy[k] = v end\n"
"        opts = copy\n"
"    end\n"
"    local host = host_or_uri\n"
"    if port == nil then\n"
"        local url = urilib.parse(tostring(host))\n"
"        if url == nil or url.service == nil then\n"
"            box.error(E_PROC_LUA,\n"
"                \"usage: connect(uri[, opts] | host, port[, opts])\")\n"
"        end\n"
"        host, port = url.host, url.service\n"
"        if opts.user == nil and opts.password == nil then\n"
"            opts.user, opts.password = url.login, url.password\n"
"        end\n"
"    end\n"
"    return host, port, opts\n"
"end\n"
"\n"
"local function remote_serialize(self)\n"
"    return {\n"
"        host = self.host,\n"
"        port = self.port,\n"
"        opts = next(self.opts) and self.opts,\n"
"        state = self.state,\n"
"        error = self.error,\n"
"        protocol = self.protocol,\n"
"        schema_version = self.schema_version,\n"
"        peer_uuid = self.peer_uuid,\n"
"        peer_version_id = self.peer_version_id\n"
"    }\n"
"end\n"
"\n"
"local remote_methods = {}\n"
"local remote_mt = {\n"
"    __index = remote_methods, __serialize = remote_serialize,\n"
"    __metatable = false\n"
"}\n"
"\n"
"local console_methods = {}\n"
"local console_mt = {\n"
"    __index = console_methods, __serialize = remote_serialize,\n"
"    __metatable = false\n"
"}\n"
"\n"
"local space_metatable, index_metatable\n"
"\n"
"local function new_sm(host, port, opts, connection, greeting)\n"
"    local user, password = opts.user, opts.password; opts.password = nil\n"
"    local last_reconnect_error\n"
"    local remote = {host = host, port = port, opts = opts, state = 'initial'}\n"
"    local function callback(what, ...)\n"
"        if what == 'state_changed' then\n"
"            local state, errno, err = ...\n"
"            local was_connected = remote._is_connected\n"
"            if state == 'active' then\n"
"                if not was_connected then\n"
"                    remote._is_connected = true\n"
"                    remote._on_connect:run(remote)\n"
"                end\n"
"            elseif state == 'error' or state == 'error_reconnect' or\n"
"                   state == 'closed' then\n"
"                if was_connected then\n"
"                    remote._is_connected = false\n"
"                    remote._on_disconnect:run(remote)\n"
"                end\n"
"            end\n"
"            remote.state, remote.error = state, err\n"
"            if state == 'error_reconnect' then\n"
"                -- Repeat the same error in verbose log only.\n"
"                -- Else the error clogs the log. See gh-3175.\n"
"                if err ~= last_reconnect_error then\n"
"                    log.warn('%s:%s: %s', host or \"\", port or \"\", err)\n"
"                    last_reconnect_error = err\n"
"                else\n"
"                    log.verbose('%s:%s: %s', host or \"\", port or \"\", err)\n"
"                end\n"
"            end\n"
"        elseif what == 'handshake' then\n"
"            local greeting = ...\n"
"            if not opts.console and greeting.protocol ~= 'Binary' then\n"
"                return E_NO_CONNECTION,\n"
"                       'Unsupported protocol: '..greeting.protocol\n"
"            end\n"
"            remote.protocol = greeting.protocol\n"
"            remote.peer_uuid = greeting.uuid\n"
"            remote.peer_version_id = greeting.version_id\n"
"        elseif what == 'will_fetch_schema' then\n"
"            return not opts.console\n"
"        elseif what == 'fetch_connect_timeout' then\n"
"            return opts.connect_timeout or DEFAULT_CONNECT_TIMEOUT\n"
"        elseif what == 'did_fetch_schema' then\n"
"            remote:_install_schema(...)\n"
"        elseif what == 'reconnect_timeout' then\n"
"            if type(opts.reconnect_after) == 'number' and\n"
"               opts.reconnect_after > 0 then\n"
"                return opts.reconnect_after\n"
"            end\n"
"        end\n"
"    end\n"
"    -- @deprecated since 1.10\n"
"    if opts.console then\n"
"        log.warn(\"Netbox console API is deprecated since 1.10, please use \"..\n"
"                 \"require('console').connect() instead\")\n"
"        setmetatable(remote, console_mt)\n"
"    else\n"
"        setmetatable(remote, remote_mt)\n"
"        -- @deprecated since 1.7.4\n"
"        remote._deadlines = setmetatable({}, {__mode = 'k'})\n"
"\n"
"        remote._space_mt = space_metatable(remote)\n"
"        remote._index_mt = index_metatable(remote)\n"
"        if opts.call_16 then\n"
"            remote.call = remote.call_16\n"
"            remote.eval = remote.eval_16\n"
"        end\n"
"    end\n"
"    remote._on_schema_reload = trigger.new(\"on_schema_reload\")\n"
"    remote._on_disconnect = trigger.new(\"on_disconnect\")\n"
"    remote._on_connect = trigger.new(\"on_connect\")\n"
"    remote._is_connected = false\n"
"    remote._transport = create_transport(host, port, user, password, callback,\n"
"                                         connection, greeting)\n"
"    remote._transport.start()\n"
"    if opts.wait_connected ~= false then\n"
"        remote._transport.wait_state('active', tonumber(opts.wait_connected))\n"
"    end\n"
"    return remote\n"
"end\n"
"\n"
"--\n"
"-- Wrap an existing connection into net.box API.\n"
"-- @param connection Connected socket.\n"
"-- @param greeting Decoded greeting, received from a server.\n"
"-- @param host Hostname to which @a connection is established.\n"
"-- @param port TCP port to which @a connection is established.\n"
"-- @param opts Options like reconnect_after, connect_timeout,\n"
"--        wait_connected, login, password, ...\n"
"--\n"
"-- @retval Net.box object.\n"
"--\n"
"local function wrap(connection, greeting, host, port, opts)\n"
"    if connection == nil or type(greeting) ~= 'table' then\n"
"        error('Usage: netbox.wrap(socket, greeting, [opts])')\n"
"    end\n"
"    opts = opts or {}\n"
"    return new_sm(host, port, opts, connection, greeting)\n"
"end\n"
"\n"
"--\n"
"-- Connect to a remote server.\n"
"-- @param uri OR host and port. URI is a string like\n"
"--        hostname:port@login:password. Host and port can be\n"
"--        passed separately with login and password in the next\n"
"--        parameter.\n"
"-- @param opts @Sa wrap().\n"
"--\n"
"-- @retval Net.box object.\n"
"--\n"
"local function connect(...)\n"
"    return new_sm(parse_connect_params(...))\n"
"end\n"
"\n"
"local function check_remote_arg(remote, method)\n"
"    if type(remote) ~= 'table' then\n"
"        local fmt = 'Use remote:%s(...) instead of remote.%s(...):'\n"
"        box.error(E_PROC_LUA, string.format(fmt, method, method))\n"
"    end\n"
"end\n"
"\n"
"local function check_call_args(args)\n"
"    if args ~= nil and type(args) ~= 'table' then\n"
"        error(\"Use remote:call(func_name, {arg1, arg2, ...}, opts) \"..\n"
"              \"instead of remote:call(func_name, arg1, arg2, ...)\")\n"
"    end\n"
"end\n"
"\n"
"local function check_eval_args(args)\n"
"    if args ~= nil and type(args) ~= 'table' then\n"
"        error(\"Use remote:eval(expression, {arg1, arg2, ...}, opts) \"..\n"
"              \"instead of remote:eval(expression, arg1, arg2, ...)\")\n"
"    end\n"
"end\n"
"\n"
"function remote_methods:close()\n"
"    check_remote_arg(self, 'close')\n"
"    self._transport.stop()\n"
"end\n"
"\n"
"function remote_methods:on_schema_reload(...)\n"
"    check_remote_arg(self, 'on_schema_reload')\n"
"    return self._on_schema_reload(...)\n"
"end\n"
"\n"
"function remote_methods:on_disconnect(...)\n"
"    check_remote_arg(self, 'on_disconnect')\n"
"    return self._on_disconnect(...)\n"
"end\n"
"\n"
"function remote_methods:on_connect(...)\n"
"    check_remote_arg(self, 'on_connect')\n"
"    return self._on_connect(...)\n"
"end\n"
"\n"
"function remote_methods:is_connected()\n"
"    check_remote_arg(self, 'is_connected')\n"
"    return self.state == 'active' or self.state == 'fetch_schema'\n"
"end\n"
"\n"
"function remote_methods:wait_connected(timeout)\n"
"    check_remote_arg(self, 'wait_connected')\n"
"    return self._transport.wait_state('active', timeout)\n"
"end\n"
"\n"
"function remote_methods:_request(method, opts, request_ctx, ...)\n"
"    local transport = self._transport\n"
"    local on_push, on_push_ctx, buffer, skip_header, deadline\n"
"    -- Extract options, set defaults, check if the request is\n"
"    -- async.\n"
"    if opts then\n"
"        buffer = opts.buffer\n"
"        skip_header = opts.skip_header\n"
"        if opts.is_async then\n"
"            if opts.on_push or opts.on_push_ctx then\n"
"                error('To handle pushes in an async request use future:pairs()')\n"
"            end\n"
"            local res, err =\n"
"                transport.perform_async_request(buffer, skip_header, method,\n"
"                                                table.insert, {}, request_ctx,\n"
"                                                ...)\n"
"            if err then\n"
"                box.error(err)\n"
"            end\n"
"            return res\n"
"        end\n"
"        if opts.timeout then\n"
"            -- conn.space:request(, { timeout = timeout })\n"
"            deadline = fiber_clock() + opts.timeout\n"
"        else\n"
"            -- conn:timeout(timeout).space:request()\n"
"            -- @deprecated since 1.7.4\n"
"            deadline = self._deadlines[fiber_self()]\n"
"        end\n"
"        on_push = opts.on_push or on_push_sync_default\n"
"        on_push_ctx = opts.on_push_ctx\n"
"    else\n"
"        deadline = self._deadlines[fiber_self()]\n"
"        on_push = on_push_sync_default\n"
"    end\n"
"    -- Execute synchronous request.\n"
"    local timeout = deadline and max(0, deadline - fiber_clock())\n"
"    if self.state ~= 'active' then\n"
"        transport.wait_state('active', timeout)\n"
"        timeout = deadline and max(0, deadline - fiber_clock())\n"
"    end\n"
"    local res, err = transport.perform_request(timeout, buffer, skip_header,\n"
"                                               method, on_push, on_push_ctx,\n"
"                                               request_ctx, ...)\n"
"    if err then\n"
"        box.error(err)\n"
"    end\n"
"    -- Try to wait until a schema is reloaded if needed.\n"
"    -- Regardless of reloading result, the main response is\n"
"    -- returned, since it does not depend on any schema things.\n"
"    if self.state == 'fetch_schema' then\n"
"        timeout = deadline and max(0, deadline - fiber_clock())\n"
"        transport.wait_state('active', timeout)\n"
"    end\n"
"    return res\n"
"end\n"
"\n"
"function remote_methods:ping(opts)\n"
"    check_remote_arg(self, 'ping')\n"
"    return (pcall(self._request, self, 'ping', opts))\n"
"end\n"
"\n"
"function remote_methods:reload_schema()\n"
"    check_remote_arg(self, 'reload_schema')\n"
"    self:ping()\n"
"end\n"
"\n"
"-- @deprecated since 1.7.4\n"
"function remote_methods:call_16(func_name, ...)\n"
"    check_remote_arg(self, 'call')\n"
"    return (self:_request('call_16', nil, nil, tostring(func_name), {...}))\n"
"end\n"
"\n"
"function remote_methods:call(func_name, args, opts)\n"
"    check_remote_arg(self, 'call')\n"
"    check_call_args(args)\n"
"    args = args or {}\n"
"    local res = self:_request('call_17', opts, nil, tostring(func_name), args)\n"
"    if type(res) ~= 'table' or opts and opts.is_async then\n"
"        return res\n"
"    end\n"
"    return unpack(res)\n"
"end\n"
"\n"
"-- @deprecated since 1.7.4\n"
"function remote_methods:eval_16(code, ...)\n"
"    check_remote_arg(self, 'eval')\n"
"    return unpack((self:_request('eval', nil, nil, code, {...})))\n"
"end\n"
"\n"
"function remote_methods:eval(code, args, opts)\n"
"    check_remote_arg(self, 'eval')\n"
"    check_eval_args(args)\n"
"    args = args or {}\n"
"    local res = self:_request('eval', opts, nil, code, args)\n"
"    if type(res) ~= 'table' or opts and opts.is_async then\n"
"        return res\n"
"    end\n"
"    return unpack(res)\n"
"end\n"
"\n"
"function remote_methods:execute(query, parameters, sql_opts, netbox_opts)\n"
"    check_remote_arg(self, \"execute\")\n"
"    if sql_opts ~= nil then\n"
"        box.error(box.error.UNSUPPORTED, \"execute\", \"options\")\n"
"    end\n"
"    return self:_request('execute', netbox_opts, nil, query, parameters or {},\n"
"                         sql_opts or {})\n"
"end\n"
"\n"
"function remote_methods:prepare(query, parameters, sql_opts, netbox_opts)\n"
"    check_remote_arg(self, \"prepare\")\n"
"    if type(query) ~= \"string\" then\n"
"        box.error(box.error.SQL_PREPARE, \"expected string as SQL statement\")\n"
"    end\n"
"    if sql_opts ~= nil then\n"
"        box.error(box.error.UNSUPPORTED, \"prepare\", \"options\")\n"
"    end\n"
"    return self:_request('prepare', netbox_opts, nil, query)\n"
"end\n"
"\n"
"function remote_methods:unprepare(query, parameters, sql_opts, netbox_opts)\n"
"    check_remote_arg(self, \"unprepare\")\n"
"    if type(query) ~= \"number\" then\n"
"        box.error(\"query id is expected to be numeric\")\n"
"    end\n"
"    if sql_opts ~= nil then\n"
"        box.error(box.error.UNSUPPORTED, \"unprepare\", \"options\")\n"
"    end\n"
"    return self:_request('unprepare', netbox_opts, nil, query, parameters or {},\n"
"                         sql_opts or {})\n"
"end\n"
"\n"
"function remote_methods:wait_state(state, timeout)\n"
"    check_remote_arg(self, 'wait_state')\n"
"    if timeout == nil then\n"
"        local deadline = self._deadlines[fiber_self()]\n"
"        timeout = deadline and max(0, deadline - fiber_clock())\n"
"    end\n"
"    return self._transport.wait_state(state, timeout)\n"
"end\n"
"\n"
"local compat_warning_said = false\n"
"\n"
"-- @deprecated since 1.7.4\n"
"function remote_methods:timeout(timeout)\n"
"    check_remote_arg(self, 'timeout')\n"
"    if not compat_warning_said then\n"
"        compat_warning_said = true\n"
"        log.warn(\"netbox:timeout(timeout) is deprecated since 1.7.4, \"..\n"
"                 \"please use space:<request>(..., {timeout = t}) instead.\")\n"
"    end\n"
"    -- Sic: this is broken by design\n"
"    self._deadlines[fiber_self()] = (timeout and fiber_clock() + timeout)\n"
"    return self\n"
"end\n"
"\n"
"function remote_methods:_install_schema(schema_version, spaces, indices,\n"
"                                        collations)\n"
"    local sl, space_mt, index_mt = {}, self._space_mt, self._index_mt\n"
"    for _, space in pairs(spaces) do\n"
"        local name = space[3]\n"
"        local id = space[1]\n"
"        local engine = space[4]\n"
"        local field_count = space[5]\n"
"        local format = space[7] or {}\n"
"        local s = {}\n"
"        if self.space ~= nil and self.space[id] ~= nil then\n"
"            s = self.space[id]\n"
"        else\n"
"            setmetatable(s, space_mt)\n"
"        end\n"
"        s.id = id\n"
"        s.name = name\n"
"        s.engine = engine\n"
"        s.field_count = field_count\n"
"        s.enabled = true\n"
"        s.index = {}\n"
"        s.temporary = false\n"
"        s._format = format\n"
"        s._format_cdata = box.internal.new_tuple_format(format)\n"
"        s.connection = self\n"
"        if #space > 5 then\n"
"            local opts = space[6]\n"
"            if type(opts) == 'table' then\n"
"                -- Tarantool >= 1.7.0\n"
"                s.temporary = not not opts.temporary\n"
"            elseif type(opts) == 'string' then\n"
"                -- Tarantool < 1.7.0\n"
"                s.temporary = string.match(opts, 'temporary') ~= nil\n"
"            end\n"
"        end\n"
"\n"
"        sl[id] = s\n"
"        sl[name] = s\n"
"    end\n"
"\n"
"    for _, index in pairs(indices) do\n"
"        local idx = {\n"
"            space   = index[1],\n"
"            id      = index[2],\n"
"            name    = index[3],\n"
"            type    = string.upper(index[4]),\n"
"            parts   = {},\n"
"        }\n"
"        local OPTS = 5\n"
"        local PARTS = 6\n"
"\n"
"        if type(index[OPTS]) == 'number' then\n"
"            idx.unique = index[OPTS] == 1\n"
"\n"
"            for k = 0, index[PARTS] - 1 do\n"
"                local pktype = index[7 + k * 2 + 1]\n"
"                local pkfield = index[7 + k * 2]\n"
"\n"
"                local pk = {\n"
"                    type = pktype,\n"
"                    fieldno = pkfield + 1\n"
"                }\n"
"                idx.parts[k + 1] = pk\n"
"            end\n"
"        else\n"
"            for k = 1, #index[PARTS] do\n"
"                local pknullable = index[PARTS][k].is_nullable or false\n"
"                local pkcollationid = index[PARTS][k].collation\n"
"                local pktype = index[PARTS][k][2] or index[PARTS][k].type\n"
"                local pkfield = index[PARTS][k][1] or index[PARTS][k].field\n"
"                -- resolve a collation name if a peer has\n"
"                -- _vcollation view\n"
"                local pkcollation = nil\n"
"                if pkcollationid ~= nil and collations ~= nil then\n"
"                    pkcollation = collations[pkcollationid + 1][2]\n"
"                end\n"
"\n"
"                local pk = {\n"
"                    type = pktype,\n"
"                    fieldno = pkfield + 1,\n"
"                    is_nullable = pknullable\n"
"                }\n"
"                if collations == nil then\n"
"                    pk.collation_id = pkcollationid\n"
"                else\n"
"                    pk.collation = pkcollation\n"
"                end\n"
"                idx.parts[k] = pk\n"
"            end\n"
"            idx.unique = not not index[OPTS].unique\n"
"        end\n"
"\n"
"        if sl[idx.space] ~= nil then\n"
"            sl[idx.space].index[idx.id] = idx\n"
"            sl[idx.space].index[idx.name] = idx\n"
"            idx.space = sl[idx.space]\n"
"            setmetatable(idx, index_mt)\n"
"        end\n"
"    end\n"
"\n"
"    self.schema_version = schema_version\n"
"    self.space = sl\n"
"    self._on_schema_reload:run(self)\n"
"end\n"
"\n"
"-- console methods\n"
"console_methods.close = remote_methods.close\n"
"console_methods.on_schema_reload = remote_methods.on_schema_reload\n"
"console_methods.on_disconnect = remote_methods.on_disconnect\n"
"console_methods.on_connect = remote_methods.on_connect\n"
"console_methods.is_connected = remote_methods.is_connected\n"
"console_methods.wait_state = remote_methods.wait_state\n"
"function console_methods:eval(line, timeout)\n"
"    check_remote_arg(self, 'eval')\n"
"    local err, res\n"
"    local transport = self._transport\n"
"    local pr = transport.perform_request\n"
"    if self.state ~= 'active' then\n"
"        local deadline = fiber_clock() + (timeout or TIMEOUT_INFINITY)\n"
"        transport.wait_state('active', timeout)\n"
"        timeout = max(0, deadline - fiber_clock())\n"
"    end\n"
"    if self.protocol == 'Binary' then\n"
"        local loader = 'return require(\"console\").eval(...)'\n"
"        res, err = pr(timeout, nil, false, 'eval', nil, nil, nil, loader,\n"
"                      {line})\n"
"    else\n"
"        assert(self.protocol == 'Lua console')\n"
"        res, err = pr(timeout, nil, false, 'inject', nil, nil, nil,\n"
"                      line..'$EOF$\\n')\n"
"    end\n"
"    if err then\n"
"        box.error(err)\n"
"    end\n"
"    return res[1] or res\n"
"end\n"
"\n"
"local function nothing_or_data(value)\n"
"    if value ~= nil then\n"
"        return value\n"
"    end\n"
"end\n"
"\n"
"space_metatable = function(remote)\n"
"    local methods = {}\n"
"\n"
"    function methods:insert(tuple, opts)\n"
"        check_space_arg(self, 'insert')\n"
"        return remote:_request('insert', opts, self._format_cdata, self.id, tuple)\n"
"    end\n"
"\n"
"    function methods:replace(tuple, opts)\n"
"        check_space_arg(self, 'replace')\n"
"        return remote:_request('replace', opts, self._format_cdata, self.id, tuple)\n"
"    end\n"
"\n"
"    function methods:select(key, opts)\n"
"        check_space_arg(self, 'select')\n"
"        return check_primary_index(self):select(key, opts)\n"
"    end\n"
"\n"
"    function methods:delete(key, opts)\n"
"        check_space_arg(self, 'delete')\n"
"        return check_primary_index(self):delete(key, opts)\n"
"    end\n"
"\n"
"    function methods:update(key, oplist, opts)\n"
"        check_space_arg(self, 'update')\n"
"        return check_primary_index(self):update(key, oplist, opts)\n"
"    end\n"
"\n"
"    function methods:upsert(key, oplist, opts)\n"
"        check_space_arg(self, 'upsert')\n"
"        return nothing_or_data(remote:_request('upsert', opts, nil, self.id, key,\n"
"                                               oplist))\n"
"    end\n"
"\n"
"    function methods:get(key, opts)\n"
"        check_space_arg(self, 'get')\n"
"        return check_primary_index(self):get(key, opts)\n"
"    end\n"
"\n"
"    function methods:format(format)\n"
"        if format == nil then\n"
"            return self._format\n"
"        else\n"
"            box.error(box.error.UNSUPPORTED, \"net.box\", \"setting space format\")\n"
"        end\n"
"    end\n"
"\n"
"    return { __index = methods, __metatable = false }\n"
"end\n"
"\n"
"index_metatable = function(remote)\n"
"    local methods = {}\n"
"\n"
"    function methods:select(key, opts)\n"
"        check_index_arg(self, 'select')\n"
"        local key_is_nil = (key == nil or\n"
"                            (type(key) == 'table' and #key == 0))\n"
"        local iterator = check_iterator_type(opts, key_is_nil)\n"
"        local offset = tonumber(opts and opts.offset) or 0\n"
"        local limit = tonumber(opts and opts.limit) or 0xFFFFFFFF\n"
"        return (remote:_request('select', opts, self.space._format_cdata,\n"
"                                self.space.id, self.id, iterator, offset,\n"
"                                limit, key))\n"
"    end\n"
"\n"
"    function methods:get(key, opts)\n"
"        check_index_arg(self, 'get')\n"
"        if opts and opts.buffer then\n"
"            error(\"index:get() doesn't support `buffer` argument\")\n"
"        end\n"
"        return nothing_or_data(remote:_request('get', opts,\n"
"                                               self.space._format_cdata,\n"
"                                               self.space.id, self.id,\n"
"                                               box.index.EQ, 0, 2, key))\n"
"    end\n"
"\n"
"    function methods:min(key, opts)\n"
"        check_index_arg(self, 'min')\n"
"        if opts and opts.buffer then\n"
"            error(\"index:min() doesn't support `buffer` argument\")\n"
"        end\n"
"        return nothing_or_data(remote:_request('min', opts,\n"
"                                               self.space._format_cdata,\n"
"                                               self.space.id, self.id,\n"
"                                               box.index.GE, 0, 1, key))\n"
"    end\n"
"\n"
"    function methods:max(key, opts)\n"
"        check_index_arg(self, 'max')\n"
"        if opts and opts.buffer then\n"
"            error(\"index:max() doesn't support `buffer` argument\")\n"
"        end\n"
"        return nothing_or_data(remote:_request('max', opts,\n"
"                                               self.space._format_cdata,\n"
"                                               self.space.id, self.id,\n"
"                                               box.index.LE, 0, 1, key))\n"
"    end\n"
"\n"
"    function methods:count(key, opts)\n"
"        check_index_arg(self, 'count')\n"
"        if opts and opts.buffer then\n"
"            error(\"index:count() doesn't support `buffer` argument\")\n"
"        end\n"
"        local code = string.format('box.space.%s.index.%s:count',\n"
"                                   self.space.name, self.name)\n"
"        return remote:_request('count', opts, nil, code, { key, opts })\n"
"    end\n"
"\n"
"    function methods:delete(key, opts)\n"
"        check_index_arg(self, 'delete')\n"
"        return nothing_or_data(remote:_request('delete', opts,\n"
"                                               self.space._format_cdata,\n"
"                                               self.space.id, self.id, key))\n"
"    end\n"
"\n"
"    function methods:update(key, oplist, opts)\n"
"        check_index_arg(self, 'update')\n"
"        return nothing_or_data(remote:_request('update', opts,\n"
"                                               self.space._format_cdata,\n"
"                                               self.space.id, self.id, key,\n"
"                                               oplist))\n"
"    end\n"
"\n"
"    return { __index = methods, __metatable = false }\n"
"end\n"
"\n"
"local this_module = {\n"
"    create_transport = create_transport,\n"
"    connect = connect,\n"
"    new = connect, -- Tarantool < 1.7.1 compatibility,\n"
"    wrap = wrap,\n"
"    establish_connection = establish_connection,\n"
"}\n"
"\n"
"function this_module.timeout(timeout, ...)\n"
"    if type(timeout) == 'table' then timeout = ... end\n"
"    if not timeout then return this_module end\n"
"    local function timed_connect(...)\n"
"        local host, port, opts = parse_connect_params(...)\n"
"        if opts.wait_connected ~= false then opts.wait_connected = timeout end\n"
"        return connect(host, port, opts)\n"
"    end\n"
"    return setmetatable({\n"
"        connect = timed_connect, new = timed_connect\n"
"    }, {__index = this_module})\n"
"end\n"
"\n"
"local function rollback()\n"
"    if rawget(box, 'rollback') ~= nil then\n"
"        -- roll back local transaction on error\n"
"        box.rollback()\n"
"    end\n"
"end\n"
"\n"
"local function handle_eval_result(status, ...)\n"
"    if not status then\n"
"        rollback()\n"
"        return box.error(E_PROC_LUA, (...))\n"
"    end\n"
"    local results = {...}\n"
"    for i = 1, select('#', ...) do\n"
"        if type(results[i]) == 'cdata' then\n"
"            results[i] = msgpack.decode(msgpack.encode(results[i]))\n"
"        end\n"
"    end\n"
"    return unpack(results)\n"
"end\n"
"\n"
"this_module.self = {\n"
"    ping = function() return true end,\n"
"    reload_schema = function() end,\n"
"    close = function() end,\n"
"    timeout = function(self) return self end,\n"
"    wait_connected = function(self) return true end,\n"
"    is_connected = function(self) return true end,\n"
"    call = function(_box, proc_name, args, opts)\n"
"        check_remote_arg(_box, 'call')\n"
"        check_call_args(args)\n"
"        args = args or {}\n"
"        proc_name = tostring(proc_name)\n"
"        local status, proc, obj = pcall(package.loaded['box.internal'].\n"
"            call_loadproc, proc_name)\n"
"        if not status then\n"
"            rollback()\n"
"            return box.error() -- re-throw\n"
"        end\n"
"        local result\n"
"        if obj ~= nil then\n"
"            return handle_eval_result(pcall(proc, obj, unpack(args)))\n"
"        else\n"
"            return handle_eval_result(pcall(proc, unpack(args)))\n"
"        end\n"
"    end,\n"
"    eval = function(_box, expr, args, opts)\n"
"        check_remote_arg(_box, 'eval')\n"
"        check_eval_args(args)\n"
"        args = args or {}\n"
"        local proc, errmsg = loadstring(expr)\n"
"        if not proc then\n"
"            proc, errmsg = loadstring(\"return \"..expr)\n"
"        end\n"
"        if not proc then\n"
"            rollback()\n"
"            return box.error(box.error.PROC_LUA, errmsg)\n"
"        end\n"
"        return handle_eval_result(pcall(proc, unpack(args)))\n"
"    end\n"
"}\n"
"\n"
"setmetatable(this_module.self, {\n"
"    __index = function(self, key)\n"
"        if key == 'space' then\n"
"            -- proxy self.space to box.space\n"
"            return require('box').space\n"
"        end\n"
"        return nil\n"
"    end\n"
"})\n"
"\n"
"package.loaded['net.box'] = this_module\n"
""
;
