const char upgrade_lua[] =
"local log = require('log')\n"
"local bit = require('bit')\n"
"local json = require('json')\n"
"\n"
"-- Guest user id - the default user\n"
"local GUEST = 0\n"
"-- Super User ID\n"
"local ADMIN = 1\n"
"-- role 'PUBLIC' is special, it's automatically granted to every user\n"
"local PUBLIC = 2\n"
"-- role 'REPLICATION'\n"
"local REPLICATION = 3\n"
"-- role 'SUPER'\n"
"-- choose a fancy id to not clash with any existing role or\n"
"-- user during upgrade\n"
"local SUPER = 31\n"
"\n"
"--------------------------------------------------------------------------------\n"
"-- Utils\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function setmap(tab)\n"
"    return setmetatable(tab, { __serialize = 'map' })\n"
"end\n"
"\n"
"local mkversion = {}\n"
"mkversion.__index = mkversion\n"
"setmetatable(mkversion, {__call = function(c, ...) return c.new(...) end})\n"
"\n"
"function mkversion.new(major, minor, patch)\n"
"    local self = setmetatable({}, mkversion)\n"
"    self.major = major\n"
"    self.minor = minor\n"
"    self.patch = patch\n"
"    self.id = bit.bor(bit.lshift(bit.bor(bit.lshift(major, 8), minor), 8), patch)\n"
"    return self\n"
"end\n"
"\n"
"function mkversion.__tostring(self)\n"
"    return string.format('%s.%s.%s', self.major, self.minor, self.patch)\n"
"end\n"
"\n"
"function mkversion.__eq(lhs, rhs)\n"
"    return lhs.id == rhs.id\n"
"end\n"
"\n"
"function mkversion.__lt(lhs, rhs)\n"
"    return lhs.id < rhs.id\n"
"end\n"
"\n"
"-- space:truncate() doesn't work with disabled triggers on __index\n"
"local function truncate(space)\n"
"    local pk = space.index[0]\n"
"    while pk:len() > 0 do\n"
"        local state, t\n"
"        for state, t in pk:pairs() do\n"
"            local key = {}\n"
"            for _k2, parts in ipairs(pk.parts) do\n"
"                table.insert(key, t[parts.fieldno])\n"
"            end\n"
"            space:delete(key)\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function foreach_system_space(cb)\n"
"    local max = box.schema.SYSTEM_ID_MAX\n"
"    for id, space in pairs(box.space) do\n"
"        -- Check for number, because box.space contains each space\n"
"        -- twice - by ID and by name. Here IDs are selected.\n"
"        -- Engine is checked to be a 'native' space, because other\n"
"        -- engines does not support DML, and does not have\n"
"        -- triggers to turn off/on. These are 'service',\n"
"        -- 'blackhole', and more may be added.\n"
"        -- When id > max system id is met, break is not done,\n"
"        -- because box.space is not an array, and it is not safe\n"
"        -- to assume all its numeric indexes are returned in\n"
"        -- ascending order.\n"
"        if type(id) == 'number' and id <= max and\n"
"           (space.engine == 'memtx' or space.engine == 'vinyl') then\n"
"            cb(space)\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function set_system_triggers(val)\n"
"    foreach_system_space(function(s) s:run_triggers(val) end)\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"-- Bootstrap\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function erase()\n"
"    foreach_system_space(function(s) truncate(s) end)\n"
"end\n"
"\n"
"local function create_sysview(source_id, target_id)\n"
"    --\n"
"    -- Create definitions for the system view, and grant\n"
"    -- privileges on system views to 'PUBLIC' role\n"
"    --\n"
"    local def = box.space._space:get(source_id):totable()\n"
"    def[1] = target_id\n"
"    def[3] = \"_v\"..def[3]:sub(2)\n"
"    def[4] = 'sysview'\n"
"    local space_def = box.space._space:get(target_id)\n"
"    if space_def == nil then\n"
"        log.info(\"create view %s...\", def[3])\n"
"        box.space._space:replace(def)\n"
"    elseif json.encode(space_def[7]) ~= json.encode(def[7]) then\n"
"        -- sync box.space._vXXX format with box.space._XXX format\n"
"        log.info(\"alter space %s set format\", def[3])\n"
"        box.space._space:update(def[1], {{ '=', 7, def[7] }})\n"
"    end\n"
"    local idefs = {}\n"
"    for _, idef in box.space._index:pairs(source_id, { iterator = 'EQ'}) do\n"
"        idef = idef:totable()\n"
"        idef[1] = target_id\n"
"        table.insert(idefs, idef)\n"
"    end\n"
"    for _, idef in ipairs(idefs) do\n"
"        if box.space._index:get({idef[1], idef[2]}) == nil then\n"
"            log.info(\"create index %s on %s\", idef[3], def[3])\n"
"            box.space._index:replace(idef)\n"
"        end\n"
"    end\n"
"    -- public can read system views\n"
"    if box.space._priv.index.primary:count({PUBLIC, 'space', target_id}) == 0 then\n"
"        log.info(\"grant read access to 'public' role for %s view\", def[3])\n"
"        box.space._priv:insert({1, PUBLIC, 'space', target_id, box.priv.R})\n"
"    end\n"
"end\n"
"\n"
"local function initial_1_7_5()\n"
"    -- stick to the following convention:\n"
"    -- prefer user id (owner id) in field #1\n"
"    -- prefer object name in field #2\n"
"    -- index on owner id is index #1\n"
"    -- index on object name is index #2\n"
"    --\n"
"\n"
"    local _schema = box.space[box.schema.SCHEMA_ID]\n"
"    local _space = box.space[box.schema.SPACE_ID]\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    local _func = box.space[box.schema.FUNC_ID]\n"
"    local _user = box.space[box.schema.USER_ID]\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    local _cluster = box.space[box.schema.CLUSTER_ID]\n"
"    local _truncate = box.space[box.schema.TRUNCATE_ID]\n"
"    local MAP = setmap({})\n"
"    local datetime = os.date(\"%Y-%m-%d %H:%M:%S\")\n"
"\n"
"    --\n"
"    -- _schema\n"
"    --\n"
"    log.info(\"create space _schema\")\n"
"    local format = {}\n"
"    format[1] = {type='string', name='key'}\n"
"    _space:insert{_schema.id, ADMIN, '_schema', 'memtx', 0, MAP, format}\n"
"    log.info(\"create index primary on _schema\")\n"
"    _index:insert{_schema.id, 0, 'primary', 'tree', { unique = true }, {{0, 'string'}}}\n"
"\n"
"    --\n"
"    -- _space\n"
"    --\n"
"    log.info(\"create space _space\")\n"
"    format = {}\n"
"    format[1] = {name='id', type='unsigned'}\n"
"    format[2] = {name='owner', type='unsigned'}\n"
"    format[3] = {name='name', type='string'}\n"
"    format[4] = {name='engine', type='string'}\n"
"    format[5] = {name='field_count', type='unsigned'}\n"
"    format[6] = {name='flags', type='map'}\n"
"    format[7] = {name='format', type='array'}\n"
"    local def = _space:insert{_space.id, ADMIN, '_space', 'memtx', 0, MAP, format}\n"
"    -- space name is unique\n"
"    log.info(\"create index primary on _space\")\n"
"    _index:insert{_space.id, 0, 'primary', 'tree', { unique = true }, {{0, 'unsigned'}}}\n"
"    log.info(\"create index owner on _space\")\n"
"    _index:insert{_space.id, 1, 'owner', 'tree', {unique = false }, {{1, 'unsigned'}}}\n"
"    log.info(\"create index index name on _space\")\n"
"    _index:insert{_space.id, 2, 'name', 'tree', { unique = true }, {{2, 'string'}}}\n"
"    create_sysview(box.schema.SPACE_ID, box.schema.VSPACE_ID)\n"
"\n"
"    --\n"
"    -- _index\n"
"    --\n"
"    log.info(\"create space _index\")\n"
"    format = {}\n"
"    format[1] = {name = 'id', type = 'unsigned'}\n"
"    format[2] = {name = 'iid', type = 'unsigned'}\n"
"    format[3] = {name = 'name', type = 'string'}\n"
"    format[4] = {name = 'type', type = 'string'}\n"
"    format[5] = {name = 'opts', type = 'map'}\n"
"    format[6] = {name = 'parts', type = 'array'}\n"
"    def = _space:insert{_index.id, ADMIN, '_index', 'memtx', 0, MAP, format}\n"
"    -- index name is unique within a space\n"
"    log.info(\"create index primary on _index\")\n"
"    _index:insert{_index.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}, {1, 'unsigned'}}}\n"
"    log.info(\"create index name on _index\")\n"
"    _index:insert{_index.id, 2, 'name', 'tree', {unique = true}, {{0, 'unsigned'}, {2, 'string'}}}\n"
"    create_sysview(box.schema.INDEX_ID, box.schema.VINDEX_ID)\n"
"\n"
"    --\n"
"    -- _func\n"
"    --\n"
"    log.info(\"create space _func\")\n"
"    format = {}\n"
"    format[1] = {name='id', type='unsigned'}\n"
"    format[2] = {name='owner', type='unsigned'}\n"
"    format[3] = {name='name', type='string'}\n"
"    format[4] = {name='setuid', type='unsigned'}\n"
"    def = _space:insert{_func.id, ADMIN, '_func', 'memtx', 0, MAP, format}\n"
"    -- function name and id are unique\n"
"    log.info(\"create index _func:primary\")\n"
"    _index:insert{_func.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}}}\n"
"    log.info(\"create index _func:owner\")\n"
"    _index:insert{_func.id, 1, 'owner', 'tree', {unique = false}, {{1, 'unsigned'}}}\n"
"    log.info(\"create index _func:name\")\n"
"    _index:insert{_func.id, 2, 'name', 'tree', {unique = true}, {{2, 'string'}}}\n"
"    create_sysview(box.schema.FUNC_ID, box.schema.VFUNC_ID)\n"
"\n"
"    --\n"
"    -- _user\n"
"    --\n"
"    log.info(\"create space _user\")\n"
"    format = {}\n"
"    format[1] = {name='id', type='unsigned'}\n"
"    format[2] = {name='owner', type='unsigned'}\n"
"    format[3] = {name='name', type='string'}\n"
"    format[4] = {name='type', type='string'}\n"
"    format[5] = {name='auth', type='map'}\n"
"    def = _space:insert{_user.id, ADMIN, '_user', 'memtx', 0, MAP, format}\n"
"    -- user name and id are unique\n"
"    log.info(\"create index _func:primary\")\n"
"    _index:insert{_user.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}}}\n"
"    log.info(\"create index _func:owner\")\n"
"    _index:insert{_user.id, 1, 'owner', 'tree', {unique = false}, {{1, 'unsigned'}}}\n"
"    log.info(\"create index _func:name\")\n"
"    _index:insert{_user.id, 2, 'name', 'tree', {unique = true}, {{2, 'string'}}}\n"
"    create_sysview(box.schema.USER_ID, box.schema.VUSER_ID)\n"
"\n"
"    --\n"
"    -- _priv\n"
"    --\n"
"    log.info(\"create space _priv\")\n"
"    format = {}\n"
"    format[1] = {name='grantor', type='unsigned'}\n"
"    format[2] = {name='grantee', type='unsigned'}\n"
"    format[3] = {name='object_type', type='string'}\n"
"    format[4] = {name='object_id', type='unsigned'}\n"
"    format[5] = {name='privilege', type='unsigned'}\n"
"    def = _space:insert{_priv.id, ADMIN, '_priv', 'memtx', 0, MAP, format}\n"
"    -- user id, object type and object id are unique\n"
"    log.info(\"create index primary on _priv\")\n"
"    _index:insert{_priv.id, 0, 'primary', 'tree', {unique = true}, {{1, 'unsigned'}, {2, 'string'}, {3, 'unsigned'}}}\n"
"    -- owner index  - to quickly find all privileges granted by a user\n"
"    log.info(\"create index owner on _priv\")\n"
"    _index:insert{_priv.id, 1, 'owner', 'tree', {unique = false}, {{0, 'unsigned'}}}\n"
"    -- object index - to quickly find all grants on a given object\n"
"    log.info(\"create index object on _priv\")\n"
"    _index:insert{_priv.id, 2, 'object', 'tree', {unique = false}, {{2, 'string'}, {3, 'unsigned'}}}\n"
"    create_sysview(box.schema.PRIV_ID, box.schema.VPRIV_ID)\n"
"\n"
"    --\n"
"    -- _cluster\n"
"    --\n"
"    log.info(\"create space _cluster\")\n"
"    format = {}\n"
"    format[1] = {name='id', type='unsigned'}\n"
"    format[2] = {name='uuid', type='string'}\n"
"    _space:insert{_cluster.id, ADMIN, '_cluster', 'memtx', 0, MAP, format}\n"
"    -- primary key: node id\n"
"    log.info(\"create index primary on _cluster\")\n"
"    _index:insert{_cluster.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}}}\n"
"    -- node uuid key: node uuid\n"
"    log.info(\"create index uuid on _cluster\")\n"
"    _index:insert{_cluster.id, 1, 'uuid', 'tree', {unique = true}, {{1, 'string'}}}\n"
"\n"
"    --\n"
"    -- _truncate\n"
"    --\n"
"    log.info(\"create space _truncate\")\n"
"    format = {}\n"
"    format[1] = {name='id', type='unsigned'}\n"
"    format[2] = {name='count', type='unsigned'}\n"
"    _space:insert{_truncate.id, ADMIN, '_truncate', 'memtx', 0, MAP, format}\n"
"    -- primary key: space id\n"
"    log.info(\"create index primary on _truncate\")\n"
"    _index:insert{_truncate.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}}}\n"
"\n"
"    --\n"
"    -- Create users\n"
"    --\n"
"    log.info(\"create user guest\")\n"
"    _user:insert{GUEST, ADMIN, 'guest', 'user', MAP}\n"
"    box.schema.user.passwd('guest', '')\n"
"    log.info(\"create user admin\")\n"
"    _user:insert{ADMIN, ADMIN, 'admin', 'user', MAP}\n"
"    log.info(\"create role public\")\n"
"    _user:insert{PUBLIC, ADMIN, 'public', 'role', MAP}\n"
"    log.info(\"create role replication\")\n"
"    _user:insert{REPLICATION, ADMIN, 'replication', 'role', MAP}\n"
"\n"
"    --\n"
"    -- Create grants\n"
"    --\n"
"    log.info(\"grant read,write,execute on universe to admin\")\n"
"    _priv:insert{ADMIN, ADMIN, 'universe', 0, box.priv.R + box.priv.W + box.priv.X}\n"
"\n"
"    -- grant role 'public' to 'guest'\n"
"    log.info(\"grant role public to guest\")\n"
"    _priv:insert{ADMIN, GUEST, 'role', PUBLIC, box.priv.X}\n"
"\n"
"    -- replication can read the entire universe\n"
"    log.info(\"grant read on universe to replication\")\n"
"    _priv:replace{ADMIN, REPLICATION, 'universe', 0, box.priv.R}\n"
"    -- replication can append to '_cluster' system space\n"
"    log.info(\"grant write on space _cluster to replication\")\n"
"    _priv:replace{ADMIN, REPLICATION, 'space', _cluster.id, box.priv.W}\n"
"\n"
"    _priv:insert{ADMIN, PUBLIC, 'space', _truncate.id, box.priv.W}\n"
"\n"
"    -- create \"box.schema.user.info\" function\n"
"    log.info('create function \"box.schema.user.info\" with setuid')\n"
"    _func:replace({1, ADMIN, 'box.schema.user.info', 1, 'LUA', '', 'function',\n"
"                  {}, 'any', 'none', 'none', false, false, true, {'LUA'},\n"
"                  MAP, '', datetime, datetime})\n"
"\n"
"    -- grant 'public' role access to 'box.schema.user.info' function\n"
"    log.info('grant execute on function \"box.schema.user.info\" to public')\n"
"    _priv:replace{ADMIN, PUBLIC, 'function', 1, box.priv.X}\n"
"\n"
"    log.info(\"set max_id to box.schema.SYSTEM_ID_MAX\")\n"
"    _schema:insert{'max_id', box.schema.SYSTEM_ID_MAX}\n"
"\n"
"    log.info(\"set schema version to 1.7.5\")\n"
"    _schema:insert({'version', 1, 7, 5})\n"
"end\n"
"\n"
"local sequence_format = {{name = 'id', type = 'unsigned'},\n"
"                         {name = 'owner', type = 'unsigned'},\n"
"                         {name = 'name', type = 'string'},\n"
"                         {name = 'step', type = 'integer'},\n"
"                         {name = 'min', type = 'integer'},\n"
"                         {name = 'max', type = 'integer'},\n"
"                         {name = 'start', type = 'integer'},\n"
"                         {name = 'cache', type = 'integer'},\n"
"                         {name = 'cycle', type = 'boolean'}}\n"
"--------------------------------------------------------------------------------\n"
"-- Tarantool 1.7.6\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function create_sequence_space()\n"
"    local _space = box.space[box.schema.SPACE_ID]\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    local _sequence = box.space[box.schema.SEQUENCE_ID]\n"
"    local _sequence_data = box.space[box.schema.SEQUENCE_DATA_ID]\n"
"    local _space_sequence = box.space[box.schema.SPACE_SEQUENCE_ID]\n"
"    local MAP = setmap({})\n"
"\n"
"    log.info(\"create space _sequence\")\n"
"    _space:insert{_sequence.id, ADMIN, '_sequence', 'memtx', 0, MAP, sequence_format}\n"
"    log.info(\"create index _sequence:primary\")\n"
"    _index:insert{_sequence.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}}}\n"
"    log.info(\"create index _sequence:owner\")\n"
"    _index:insert{_sequence.id, 1, 'owner', 'tree', {unique = false}, {{1, 'unsigned'}}}\n"
"    log.info(\"create index _sequence:name\")\n"
"    _index:insert{_sequence.id, 2, 'name', 'tree', {unique = true}, {{2, 'string'}}}\n"
"\n"
"    log.info(\"create space _sequence_data\")\n"
"    _space:insert{_sequence_data.id, ADMIN, '_sequence_data', 'memtx', 0, MAP,\n"
"                  {{name = 'id', type = 'unsigned'}, {name = 'value', type = 'integer'}}}\n"
"    log.info(\"create index primary on _sequence_data\")\n"
"    _index:insert{_sequence_data.id, 0, 'primary', 'hash', {unique = true}, {{0, 'unsigned'}}}\n"
"\n"
"    log.info(\"create space _space_sequence\")\n"
"    _space:insert{_space_sequence.id, ADMIN, '_space_sequence', 'memtx', 0, MAP,\n"
"                  {{name = 'id', type = 'unsigned'},\n"
"                   {name = 'sequence_id', type = 'unsigned'},\n"
"                   {name = 'is_generated', type = 'boolean'}}}\n"
"    log.info(\"create index _space_sequence:primary\")\n"
"    _index:insert{_space_sequence.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}}}\n"
"    log.info(\"create index _space_sequence:sequence\")\n"
"    _index:insert{_space_sequence.id, 1, 'sequence', 'tree', {unique = false}, {{1, 'unsigned'}}}\n"
"end\n"
"\n"
"local function create_collation_space()\n"
"    local _collation = box.space[box.schema.COLLATION_ID]\n"
"\n"
"    log.info(\"create space _collation\")\n"
"    box.space._space:insert{_collation.id, ADMIN, '_collation', 'memtx', 0, setmap({}),\n"
"        { { name = 'id', type = 'unsigned' }, { name = 'name', type = 'string' },\n"
"          { name = 'owner', type = 'unsigned' }, { name = 'type', type = 'string' },\n"
"          { name = 'locale', type = 'string' }, { name = 'opts', type = 'map' } } }\n"
"\n"
"    log.info(\"create index primary on _collation\")\n"
"    box.space._index:insert{_collation.id, 0, 'primary', 'tree', {unique = true}, {{0, 'unsigned'}}}\n"
"\n"
"    log.info(\"create index name on _collation\")\n"
"    box.space._index:insert{_collation.id, 1, 'name', 'tree', {unique = true}, {{1, 'string'}}}\n"
"\n"
"    log.info(\"create predefined collations\")\n"
"    box.space._collation:replace{1, \"unicode\", ADMIN, \"ICU\", \"\", setmap{}}\n"
"    box.space._collation:replace{2, \"unicode_ci\", ADMIN, \"ICU\", \"\", {strength='primary'}}\n"
"\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    _priv:insert{ADMIN, PUBLIC, 'space', _collation.id, box.priv.W}\n"
"end\n"
"\n"
"local function upgrade_to_1_7_6()\n"
"    create_sequence_space()\n"
"    create_collation_space()\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"--- Tarantool 1.7.7\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function upgrade_to_1_7_7()\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    local _user = box.space[box.schema.USER_ID]\n"
"    --\n"
"    -- grant 'session' and 'usage' to all existing users\n"
"    --\n"
"    for _, v in _user:pairs() do\n"
"        if v[4] ~= \"role\" then\n"
"            _priv:upsert({ADMIN, v[1], \"universe\", 0, box.priv.S + box.priv.U},\n"
"                                                {{\"|\", 5, box.priv.S + box.priv.U}})\n"
"        end\n"
"    end\n"
"    --\n"
"    -- grant 'create' to all users with 'read' and 'write'\n"
"    -- on the universe, since going forward we will require\n"
"    -- 'create' rather than 'read,write' to be able to create\n"
"    -- objects\n"
"    --\n"
"    for _, v in _priv.index.object:pairs{'universe'} do\n"
"        if bit.band(v[5], 1) ~= 0 and bit.band(v[5], 2) ~= 0 then\n"
"            _priv:update({v[2], v[3], v[4]}, {{ \"|\", 5, box.priv.C}})\n"
"        end\n"
"    end\n"
"    -- grant admin all new privileges (session, usage, grant option,\n"
"    -- create, alter, drop and anything that might come up in the future\n"
"    --\n"
"    _priv:upsert({ADMIN, ADMIN, 'universe', 0, box.priv.ALL},\n"
"                 {{ \"|\", 5, box.priv.ALL}})\n"
"    --\n"
"    -- create role 'super' and grant it all privileges on universe\n"
"    --\n"
"    _user:replace{SUPER, ADMIN, 'super', 'role', setmap({})}\n"
"    _priv:replace({ADMIN, SUPER, 'universe', 0, 4294967295})\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"--- Tarantool 1.10.0\n"
"--------------------------------------------------------------------------------\n"
"local function create_vsequence_space()\n"
"    create_sysview(box.schema.SEQUENCE_ID, box.schema.VSEQUENCE_ID)\n"
"    box.space._vsequence:format(sequence_format)\n"
"end\n"
"\n"
"local function upgrade_to_1_10_0()\n"
"    create_vsequence_space()\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"--- Tarantool 1.10.2\n"
"--------------------------------------------------------------------------------\n"
"local function upgrade_priv_to_1_10_2()\n"
"    local _priv = box.space._priv\n"
"    local _vpriv = box.space._vpriv\n"
"    local format = _priv:format()\n"
"\n"
"    format[4].type = 'scalar'\n"
"    _priv:format(format)\n"
"    format = _vpriv:format()\n"
"    format[4].type = 'scalar'\n"
"    _vpriv:format(format)\n"
"    _priv.index.primary:alter{parts={2, 'unsigned', 3, 'string', 4, 'scalar'}}\n"
"    _vpriv.index.primary:alter{parts={2, 'unsigned', 3, 'string', 4, 'scalar'}}\n"
"    _priv.index.object:alter{parts={3, 'string', 4, 'scalar'}}\n"
"    _vpriv.index.object:alter{parts={3, 'string', 4, 'scalar'}}\n"
"end\n"
"\n"
"local function create_vinyl_deferred_delete_space()\n"
"    local _space = box.space[box.schema.SPACE_ID]\n"
"    local _vinyl_deferred_delete = box.space[box.schema.VINYL_DEFERRED_DELETE_ID]\n"
"\n"
"    local format = {}\n"
"    format[1] = {name = 'space_id', type = 'unsigned'}\n"
"    format[2] = {name = 'lsn', type = 'unsigned'}\n"
"    format[3] = {name = 'tuple', type = 'array'}\n"
"\n"
"    log.info(\"create space _vinyl_deferred_delete\")\n"
"    _space:insert{_vinyl_deferred_delete.id, ADMIN, '_vinyl_deferred_delete',\n"
"                  'blackhole', 0, {group_id = 1}, format}\n"
"end\n"
"\n"
"local function upgrade_to_1_10_2()\n"
"    upgrade_priv_to_1_10_2()\n"
"    create_vinyl_deferred_delete_space()\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"-- Tarantool 2.1.0\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function upgrade_priv_to_2_1_0()\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    local _user = box.space[box.schema.USER_ID]\n"
"    -- Since we remove 1.7 compatibility in 2.1.0, we have to\n"
"    -- grant ALTER and DROP to all users with READ + WRITE on\n"
"    -- respective objects. We also grant CREATE on entities\n"
"    -- or on universe if a user has READ and WRITE on an entity\n"
"    -- or on universe respectively. We do not grant CREATE on\n"
"    -- objects, since it has no effect. We also skip grants for\n"
"    -- sequences since they were added after the new privileges\n"
"    -- and compatibility mode was always off for them.\n"
"    for _, user in _user:pairs() do\n"
"        if user[0] ~= ADMIN and user[0] ~= SUPER then\n"
"            for _, priv in _priv:pairs(user[0]) do\n"
"                if priv[3] ~= 'sequence' and\n"
"                   bit.band(priv[5], box.priv.W) ~= 0 and\n"
"                   bit.band(priv[5], box.priv.R) ~= 0 then\n"
"                    local new_privs = bit.bor(box.priv.A, box.priv.D)\n"
"                    if priv[3] == 'universe' or priv[4] == '' then\n"
"                        new_privs = bit.bor(new_privs, box.priv.C)\n"
"                    end\n"
"                    _priv:update({priv[2], priv[3], priv[4]},\n"
"                                 {{\"|\", 5, new_privs}})\n"
"                end\n"
"            end\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function upgrade_to_2_1_0()\n"
"    local _space = box.space[box.schema.SPACE_ID]\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    local _trigger = box.space[box.schema.TRIGGER_ID]\n"
"    local MAP = setmap({})\n"
"\n"
"    log.info(\"create space _trigger\")\n"
"    local format = {{name='name', type='string'},\n"
"                    {name='space_id', type='unsigned'},\n"
"                    {name='opts', type='map'}}\n"
"    _space:insert{_trigger.id, ADMIN, '_trigger', 'memtx', 0, MAP, format}\n"
"\n"
"    log.info(\"create index primary on _trigger\")\n"
"    _index:insert{_trigger.id, 0, 'primary', 'tree', { unique = true },\n"
"                  {{0, 'string'}}}\n"
"    log.info(\"create index secondary on _trigger\")\n"
"    _index:insert{_trigger.id, 1, 'space_id', 'tree', { unique = false },\n"
"                  {{1, 'unsigned'}}}\n"
"\n"
"    local fk_constr_ft = {{name='name', type='string'},\n"
"                          {name='child_id', type='unsigned'},\n"
"                          {name='parent_id', type='unsigned'},\n"
"                          {name='is_deferred', type='boolean'},\n"
"                          {name='match', type='string'},\n"
"                          {name='on_delete', type='string'},\n"
"                          {name='on_update', type='string'},\n"
"                          {name='child_cols', type='array'},\n"
"                          {name='parent_cols', type='array'}}\n"
"    log.info(\"create space _fk_constraint\")\n"
"    _space:insert{box.schema.FK_CONSTRAINT_ID, ADMIN, '_fk_constraint', 'memtx',\n"
"                  0, setmap({}), fk_constr_ft}\n"
"\n"
"    log.info(\"create index primary on _fk_constraint\")\n"
"    _index:insert{box.schema.FK_CONSTRAINT_ID, 0, 'primary', 'tree',\n"
"                  {unique = true}, {{0, 'string'}, {1, 'unsigned'}}}\n"
"\n"
"    log.info(\"create secondary index child_id on _fk_constraint\")\n"
"    _index:insert{box.schema.FK_CONSTRAINT_ID, 1, 'child_id', 'tree',\n"
"                  {unique = false}, {{1, 'unsigned'}}}\n"
"\n"
"    -- Nullability wasn't skipable. This was fixed in 1-7.\n"
"    -- Now, abscent field means NULL, so we can safely set second\n"
"    -- field in format, marking it nullable.\n"
"    log.info(\"Add nullable value field to space _schema\")\n"
"    local format = {}\n"
"    format[1] = {type='string', name='key'}\n"
"    format[2] = {type='any', name='value', is_nullable=true}\n"
"    box.space._schema:format(format)\n"
"\n"
"    box.space._collation:replace{0, \"none\", ADMIN, \"BINARY\", \"\", setmap{}}\n"
"    box.space._collation:replace{3, \"binary\", ADMIN, \"BINARY\", \"\", setmap{}}\n"
"\n"
"    upgrade_priv_to_2_1_0()\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"-- Tarantool 2.1.1\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function upgrade_to_2_1_1()\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    for _, index in _index:pairs() do\n"
"        local opts = index.opts\n"
"        if opts['sql'] ~= nil then\n"
"            opts['sql'] = nil\n"
"            _index:replace(box.tuple.new({index.id, index.iid, index.name,\n"
"                                        index.type, opts, index.parts}))\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"-- Tarantool 2.1.2\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function update_collation_strength_field()\n"
"    local _collation = box.space[box.schema.COLLATION_ID]\n"
"    for _, collation in _collation:pairs() do\n"
"        if collation.type == 'ICU' and collation.opts.strength == nil then\n"
"            local new_collation = collation:totable()\n"
"            new_collation[6].strength = 'tertiary'\n"
"            _collation:delete{collation.id}\n"
"            _collation:insert(new_collation)\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function upgrade_to_2_1_2()\n"
"    update_collation_strength_field()\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"-- Tarantool 2.1.3\n"
"--------------------------------------------------------------------------------\n"
"\n"
"-- Add new collations\n"
"local function upgrade_collation_to_2_1_3()\n"
"    local coll_lst = {\n"
"        {name=\"af\", loc_str=\"af\"},  -- Afrikaans\n"
"        {name=\"am\", loc_str=\"am\"},  -- Amharic (no character changes, just re-ordering)\n"
"        {name=\"ar\", loc_str=\"ar\"},  -- Arabic (use only \"standard\")\n"
"        {name=\"as\", loc_str=\"as\"},  -- Assamese\n"
"        {name=\"az\", loc_str=\"az\"},  -- Azerbaijani (Azeri)\n"
"        {name=\"be\", loc_str=\"be\"},  -- Belarusian\n"
"        {name=\"bn\", loc_str=\"bn\"},  -- Bengali (Bangla actually)\n"
"        {name=\"bs\", loc_str=\"bs\"},  -- Bosnian (tailored as Croatian)\n"
"        {name=\"bs_Cyrl\", loc_str=\"bs_Cyrl\"}, -- Bosnian in Cyrillic (tailored as Serbian)\n"
"        {name=\"ca\", loc_str=\"ca\"},  -- Catalan\n"
"        {name=\"cs\", loc_str=\"cs\"},  -- Czech\n"
"        {name=\"cy\", loc_str=\"cy\"},  -- Welsh\n"
"        {name=\"da\", loc_str=\"da\"},  -- Danish\n"
"        {name=\"de__phonebook\", loc_str=\"de_DE_u_co_phonebk\"}, -- German (umlaut as 'ae', 'oe', 'ue')\n"
"        {name=\"de_AT_phonebook\", loc_str=\"de_AT_u_co_phonebk\"}, -- Austrian German (umlaut primary greater)\n"
"        {name=\"dsb\", loc_str=\"dsb\"}, -- Lower Sorbian\n"
"        {name=\"ee\", loc_str=\"ee\"},  -- Ewe\n"
"        {name=\"eo\", loc_str=\"eo\"},  -- Esperanto\n"
"        {name=\"es\", loc_str=\"es\"},  -- Spanish\n"
"        {name=\"es__traditional\", loc_str=\"es_u_co_trad\"}, -- Spanish ('ch' and 'll' as a grapheme)\n"
"        {name=\"et\", loc_str=\"et\"},  -- Estonian\n"
"        {name=\"fa\", loc_str=\"fa\"},  -- Persian\n"
"        {name=\"fi\", loc_str=\"fi\"},  -- Finnish (v and w are primary equal)\n"
"        {name=\"fi__phonebook\", loc_str=\"fi_u_co_phonebk\"}, -- Finnish (v and w as separate characters)\n"
"        {name=\"fil\", loc_str=\"fil\"}, -- Filipino\n"
"        {name=\"fo\", loc_str=\"fo\"},  -- Faroese\n"
"        {name=\"fr_CA\", loc_str=\"fr_CA\"}, -- Canadian French\n"
"        {name=\"gu\", loc_str=\"gu\"},  -- Gujarati\n"
"        {name=\"ha\", loc_str=\"ha\"},  -- Hausa\n"
"        {name=\"haw\", loc_str=\"haw\"}, -- Hawaiian\n"
"        {name=\"he\", loc_str=\"he\"},  -- Hebrew\n"
"        {name=\"hi\", loc_str=\"hi\"},  -- Hindi\n"
"        {name=\"hr\", loc_str=\"hr\"},  -- Croatian\n"
"        {name=\"hu\", loc_str=\"hu\"},  -- Hungarian\n"
"        {name=\"hy\", loc_str=\"hy\"},  -- Armenian\n"
"        {name=\"ig\", loc_str=\"ig\"},  -- Igbo\n"
"        {name=\"is\", loc_str=\"is\"},  -- Icelandic\n"
"        {name=\"ja\", loc_str=\"ja\"},  -- Japanese\n"
"        {name=\"kk\", loc_str=\"kk\"},  -- Kazakh\n"
"        {name=\"kl\", loc_str=\"kl\"},  -- Kalaallisut\n"
"        {name=\"kn\", loc_str=\"kn\"},  -- Kannada\n"
"        {name=\"ko\", loc_str=\"ko\"},  -- Korean\n"
"        {name=\"kok\", loc_str=\"kok\"}, -- Konkani\n"
"        {name=\"ky\", loc_str=\"ky\"},  -- Kyrgyz\n"
"        {name=\"lkt\", loc_str=\"lkt\"}, -- Lakota\n"
"        {name=\"ln\", loc_str=\"ln\"},  -- Lingala\n"
"        {name=\"lt\", loc_str=\"lt\"},  -- Lithuanian\n"
"        {name=\"lv\", loc_str=\"lv\"},  -- Latvian\n"
"        {name=\"mk\", loc_str=\"mk\"},  -- Macedonian\n"
"        {name=\"ml\", loc_str=\"ml\"},  -- Malayalam\n"
"        {name=\"mr\", loc_str=\"mr\"},  -- Marathi\n"
"        {name=\"mt\", loc_str=\"mt\"},  -- Maltese\n"
"        {name=\"nb\", loc_str=\"nb\"},  -- Norwegian Bokmal\n"
"        {name=\"nn\", loc_str=\"nn\"},  -- Norwegian Nynorsk\n"
"        {name=\"nso\", loc_str=\"nso\"}, -- Northern Sotho\n"
"        {name=\"om\", loc_str=\"om\"},  -- Oromo\n"
"        {name=\"or\", loc_str=\"or\"},  -- Oriya (Odia)\n"
"        {name=\"pa\", loc_str=\"pa\"},  -- Punjabi\n"
"        {name=\"pl\", loc_str=\"pl\"},  -- Polish\n"
"        {name=\"ro\", loc_str=\"ro\"},  -- Romanian\n"
"        {name=\"sa\", loc_str=\"sa\"},  -- Sanskrit\n"
"        {name=\"se\", loc_str=\"se\"},  -- Northern Sami\n"
"        {name=\"si\", loc_str=\"si\"},  -- Sinhala\n"
"        {name=\"si__dictionary\", loc_str=\"si_u_co_dict\"}, -- Sinhala (U+0DA5 = U+0DA2,0DCA,0DA4)\n"
"        {name=\"sk\", loc_str=\"sk\"},  -- Slovak\n"
"        {name=\"sl\", loc_str=\"sl\"},  -- Slovenian\n"
"        {name=\"sq\", loc_str=\"sq\"},  -- Albanian (just \"standard\")\n"
"        {name=\"sr\", loc_str=\"sr\"},  -- Serbian\n"
"        {name=\"sr_Latn\", loc_str=\"sr_Latn\"}, -- Serbian in Latin (tailored as Croatian)\n"
"        {name=\"sv\", loc_str=\"sv\"},  -- Swedish (v and w are primary equal)\n"
"        {name=\"sv__reformed\", loc_str=\"sv_u_co_reformed\"}, -- Swedish (v and w as separate characters)\n"
"        {name=\"ta\", loc_str=\"ta\"},  -- Tamil\n"
"        {name=\"te\", loc_str=\"te\"},  -- Telugu\n"
"        {name=\"th\", loc_str=\"th\"},  -- Thai\n"
"        {name=\"tn\", loc_str=\"tn\"},  -- Tswana\n"
"        {name=\"to\", loc_str=\"to\"},  -- Tonga\n"
"        {name=\"tr\", loc_str=\"tr\"},  -- Turkish\n"
"        {name=\"ug_Cyrl\", loc_str=\"ug\"}, -- Uyghur in Cyrillic - is there such locale\?\n"
"        {name=\"uk\", loc_str=\"uk\"},  -- Ukrainian\n"
"        {name=\"ur\", loc_str=\"ur\"},  -- Urdu\n"
"        {name=\"vi\", loc_str=\"vi\"},  -- Vietnamese\n"
"        {name=\"vo\", loc_str=\"vo\"},  -- Volapük\n"
"        {name=\"wae\", loc_str=\"wae\"}, -- Walser\n"
"        {name=\"wo\", loc_str=\"wo\"},  -- Wolof\n"
"        {name=\"yo\", loc_str=\"yo\"},  -- Yoruba\n"
"        {name=\"zh\", loc_str=\"zh\"},  -- Chinese\n"
"        {name=\"zh__big5han\", loc_str=\"zh_u_co_big5han\"},  -- Chinese (ideographs: big5 order)\n"
"        {name=\"zh__gb2312han\", loc_str=\"zh_u_co_gb2312\"}, -- Chinese (ideographs: GB-2312 order)\n"
"        {name=\"zh__pinyin\", loc_str=\"zh_u_co_pinyin\"}, -- Chinese (ideographs: pinyin order)\n"
"        {name=\"zh__stroke\", loc_str=\"zh_u_co_stroke\"}, -- Chinese (ideographs: stroke order)\n"
"        {name=\"zh__zhuyin\", loc_str=\"zh_u_co_zhuyin\"}, -- Chinese (ideographs: zhuyin order)\n"
"    }\n"
"    local coll_strengths = {\n"
"        {s=\"s1\", opt={strength='primary'}},\n"
"        {s=\"s2\", opt={strength='secondary'}},\n"
"        {s=\"s3\", opt={strength='tertiary'}}\n"
"    }\n"
"\n"
"    local id = 4\n"
"    for _, collation in ipairs(coll_lst) do\n"
"        for i, strength in ipairs(coll_strengths) do\n"
"            local coll_name = 'unicode_' .. collation.name .. \"_\" .. strength.s\n"
"            log.info(\"creating collation %s\", coll_name)\n"
"            box.space._collation:replace{id, coll_name, ADMIN, \"ICU\", collation.loc_str, strength.opt }\n"
"            id = id + 1\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function upgrade_to_2_1_3()\n"
"    upgrade_collation_to_2_1_3()\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"-- Tarantool 2.2.1\n"
"--------------------------------------------------------------------------------\n"
"\n"
"-- Add sequence field to _space_sequence table\n"
"local function upgrade_sequence_to_2_2_1()\n"
"    log.info(\"add sequence field to space _space_sequence\")\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    local _space_sequence = box.space[box.schema.SPACE_SEQUENCE_ID]\n"
"    for _, v in _space_sequence:pairs() do\n"
"        if #v > 3 then\n"
"            -- Must be a sequence created after upgrade.\n"
"            -- It doesn't need to get updated.\n"
"            goto continue\n"
"        end\n"
"        -- Explicitly attach the sequence to the first index part.\n"
"        local pk = _index:get{v[1], 0}\n"
"        local part = pk[6][1]\n"
"        local field = part.field or part[1]\n"
"        local path = part.path or ''\n"
"        local t = _space_sequence:get(v[1])\n"
"        -- Update in-place is banned due to complexity of its\n"
"        -- handling. Delete + insert still work.\n"
"        t = t:update({{'!', 4, field}, {'!', 5, path}})\n"
"        _space_sequence:delete({v[1]})\n"
"        _space_sequence:insert(t)\n"
"        ::continue::\n"
"    end\n"
"    local format = _space_sequence:format()\n"
"    format[4] = {name = 'field', type = 'unsigned'}\n"
"    format[5] = {name = 'path', type = 'string'}\n"
"    _space_sequence:format(format)\n"
"end\n"
"\n"
"local function upgrade_ck_constraint_to_2_2_1()\n"
"    -- In previous Tarantool releases check constraints were\n"
"    -- stored in space opts. Now we use separate space\n"
"    -- _ck_constraint for this purpose. Perform legacy data\n"
"    -- migration.\n"
"    local MAP = setmap({})\n"
"    local _space = box.space._space\n"
"    local _index = box.space._index\n"
"    local _ck_constraint = box.space._ck_constraint\n"
"    log.info(\"create space _ck_constraint\")\n"
"    local format = {{name='space_id', type='unsigned'},\n"
"                    {name='name', type='string'},\n"
"                    {name='is_deferred', type='boolean'},\n"
"                    {name='language', type='str'}, {name='code', type='str'}}\n"
"    _space:insert{_ck_constraint.id, ADMIN, '_ck_constraint', 'memtx', 0, MAP, format}\n"
"\n"
"    log.info(\"create index primary on _ck_constraint\")\n"
"    _index:insert{_ck_constraint.id, 0, 'primary', 'tree',\n"
"                  {unique = true}, {{0, 'unsigned'}, {1, 'string'}}}\n"
"\n"
"    for _, space in _space:pairs() do\n"
"        local flags = space.flags\n"
"        if flags.checks then\n"
"            for i, check in pairs(flags.checks) do\n"
"                local expr_str = check.expr\n"
"                local check_name = check.name or\n"
"                                   \"CK_CONSTRAINT_\"..i..\"_\"..space.name\n"
"                _ck_constraint:insert({space.id, check_name, false,\n"
"                                       'SQL', expr_str})\n"
"            end\n"
"            flags.checks = nil\n"
"            _space:replace({space.id, space.owner, space.name, space.engine,\n"
"                            space.field_count, flags, space.format})\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function create_vcollation_space()\n"
"    local _collation = box.space._collation\n"
"    local format = _collation:format()\n"
"    create_sysview(box.schema.COLLATION_ID, box.schema.VCOLLATION_ID)\n"
"    box.space[box.schema.VCOLLATION_ID]:format(format)\n"
"end\n"
"\n"
"local function upgrade_func_to_2_2_1()\n"
"    log.info(\"Update _func format\")\n"
"    local _func = box.space[box.schema.FUNC_ID]\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    local datetime = os.date(\"%Y-%m-%d %H:%M:%S\")\n"
"    for _, v in box.space._func:pairs() do\n"
"        box.space._func:replace({v.id, v.owner, v.name, v.setuid, v[5] or 'LUA',\n"
"                                 '', 'function', {}, 'any', 'none', 'none',\n"
"                                 false, false, true, v[15] or {'LUA'},\n"
"                                 setmap({}), '', datetime, datetime})\n"
"    end\n"
"    local sql_builtin_list = {\n"
"        \"TRIM\", \"TYPEOF\", \"PRINTF\", \"UNICODE\", \"CHAR\", \"HEX\", \"VERSION\",\n"
"        \"QUOTE\", \"REPLACE\", \"SUBSTR\", \"GROUP_CONCAT\", \"JULIANDAY\", \"DATE\",\n"
"        \"TIME\", \"DATETIME\", \"STRFTIME\", \"CURRENT_TIME\", \"CURRENT_TIMESTAMP\",\n"
"        \"CURRENT_DATE\", \"LENGTH\", \"POSITION\", \"ROUND\", \"UPPER\", \"LOWER\",\n"
"        \"IFNULL\", \"RANDOM\", \"CEIL\", \"CEILING\", \"CHARACTER_LENGTH\",\n"
"        \"CHAR_LENGTH\", \"FLOOR\", \"MOD\", \"OCTET_LENGTH\", \"ROW_COUNT\", \"COUNT\",\n"
"        \"LIKE\", \"ABS\", \"EXP\", \"LN\", \"POWER\", \"SQRT\", \"SUM\", \"TOTAL\", \"AVG\",\n"
"        \"RANDOMBLOB\", \"NULLIF\", \"ZEROBLOB\", \"MIN\", \"MAX\", \"COALESCE\", \"EVERY\",\n"
"        \"EXISTS\", \"EXTRACT\", \"SOME\", \"GREATER\", \"LESSER\", \"SOUNDEX\",\n"
"        \"LIKELIHOOD\", \"LIKELY\", \"UNLIKELY\", \"_sql_stat_get\", \"_sql_stat_push\",\n"
"        \"_sql_stat_init\",\n"
"    }\n"
"    for _, v in pairs(sql_builtin_list) do\n"
"        local t = _func:auto_increment({ADMIN, v, 1, 'SQL_BUILTIN', '',\n"
"                                       'function', {}, 'any', 'none', 'none',\n"
"                                        false, false, true, {}, setmap({}), '',\n"
"                                        datetime, datetime})\n"
"        _priv:replace{ADMIN, PUBLIC, 'function', t.id, box.priv.X}\n"
"    end\n"
"    local t = _func:auto_increment({ADMIN, 'LUA', 1, 'LUA',\n"
"                        'function(code) return assert(loadstring(code))() end',\n"
"                        'function', {'string'}, 'any', 'none', 'none',\n"
"                        false, false, true, {'LUA', 'SQL'},\n"
"                        setmap({}), '', datetime, datetime})\n"
"    _priv:replace{ADMIN, PUBLIC, 'function', t.id, box.priv.X}\n"
"    local format = {}\n"
"    format[1] = {name='id', type='unsigned'}\n"
"    format[2] = {name='owner', type='unsigned'}\n"
"    format[3] = {name='name', type='string'}\n"
"    format[4] = {name='setuid', type='unsigned'}\n"
"    format[5] = {name='language', type='string'}\n"
"    format[6] = {name='body', type='string'}\n"
"    format[7] = {name='routine_type', type='string'}\n"
"    format[8] = {name='param_list', type='array'}\n"
"    format[9] = {name='returns', type='string'}\n"
"    format[10] = {name='aggregate', type='string'}\n"
"    format[11] = {name='sql_data_access', type='string'}\n"
"    format[12] = {name='is_deterministic', type='boolean'}\n"
"    format[13] = {name='is_sandboxed', type='boolean'}\n"
"    format[14] = {name='is_null_call', type='boolean'}\n"
"    format[15] = {name='exports', type='array'}\n"
"    format[16] = {name='opts', type='map'}\n"
"    format[17] = {name='comment', type='string'}\n"
"    format[18] = {name='created', type='string'}\n"
"    format[19] = {name='last_altered', type='string'}\n"
"    _func:format(format)\n"
"    _func.index.name:alter({parts = {{'name', 'string',\n"
"                                      collation = 'unicode_ci'}}})\n"
"end\n"
"\n"
"local function create_func_index()\n"
"    log.info(\"Create _func_index space\")\n"
"    local _func_index = box.space[box.schema.FUNC_INDEX_ID]\n"
"    local _space = box.space._space\n"
"    local _index = box.space._index\n"
"    local format = {{name='space_id', type='unsigned'},\n"
"                    {name='index_id', type='unsigned'},\n"
"                    {name='func_id',  type='unsigned'}}\n"
"    _space:insert{_func_index.id, ADMIN, '_func_index', 'memtx', 0,\n"
"                  setmap({}), format}\n"
"    _index:insert{_func_index.id, 0, 'primary', 'tree', {unique = true},\n"
"                  {{0, 'unsigned'}, {1, 'unsigned'}}}\n"
"    _index:insert{_func_index.id, 1, 'fid', 'tree', {unique = false},\n"
"                  {{2, 'unsigned'}}}\n"
"\n"
"end\n"
"\n"
"local function upgrade_to_2_2_1()\n"
"    upgrade_sequence_to_2_2_1()\n"
"    upgrade_ck_constraint_to_2_2_1()\n"
"    create_vcollation_space()\n"
"    upgrade_func_to_2_2_1()\n"
"    create_func_index()\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"-- Tarantool 2.3.0\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function upgrade_to_2_3_0()\n"
"    log.info(\"Create GREATEST and LEAST SQL Builtins\")\n"
"    local _func = box.space[box.schema.FUNC_ID]\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    local datetime = os.date(\"%Y-%m-%d %H:%M:%S\")\n"
"    local new_builtins = {\"GREATEST\", \"LEAST\"}\n"
"    for _, v in pairs(new_builtins) do\n"
"        local t = _func:auto_increment({ADMIN, v, 1, 'SQL_BUILTIN', '',\n"
"                                       'function', {}, 'any', 'none', 'none',\n"
"                                        false, false, true, {}, setmap({}), '',\n"
"                                        datetime, datetime})\n"
"        _priv:replace{ADMIN, PUBLIC, 'function', t.id, box.priv.X}\n"
"    end\n"
"\n"
"    log.info(\"Extend _ck_constraint space format with is_enabled field\")\n"
"    local _ck_constraint = box.space._ck_constraint\n"
"    for _, tuple in _ck_constraint:pairs() do\n"
"        _ck_constraint:update({tuple[1], tuple[2]}, {{'=', 6, true}})\n"
"    end\n"
"    local format = {{name='space_id', type='unsigned'},\n"
"                    {name='name', type='string'},\n"
"                    {name='is_deferred', type='boolean'},\n"
"                    {name='language', type='str'},\n"
"                    {name='code', type='str'},\n"
"                    {name='is_enabled', type='boolean'}}\n"
"    _ck_constraint:format(format)\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"-- Tarantool 2.3.1\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function drop_func_collation()\n"
"    local _func = box.space[box.schema.FUNC_ID]\n"
"    _func.index.name:alter({parts = {{'name', 'string'}}})\n"
"end\n"
"\n"
"local function create_session_settings_space()\n"
"    local _space = box.space[box.schema.SPACE_ID]\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    local format = {}\n"
"    format[1] = {name='name', type='string'}\n"
"    format[2] = {name='value', type='any'}\n"
"    log.info(\"create space _session_settings\")\n"
"    _space:insert{box.schema.SESSION_SETTINGS_ID, ADMIN, '_session_settings',\n"
"                  'service', 2, {temporary = true}, format}\n"
"    log.info(\"create index _session_settings:primary\")\n"
"    _index:insert{box.schema.SESSION_SETTINGS_ID, 0, 'primary', 'tree',\n"
"                  {unique = true}, {{0, 'string'}}}\n"
"end\n"
"\n"
"local function upgrade_to_2_3_1()\n"
"    drop_func_collation()\n"
"    create_session_settings_space()\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local function get_version()\n"
"    local version = box.space._schema:get{'version'}\n"
"    if version == nil then\n"
"        error('Missing \"version\" in box.space._schema')\n"
"    end\n"
"    local major = version[2]\n"
"    local minor = version[3]\n"
"    local patch = version[4] or 0\n"
"\n"
"    return mkversion(major, minor, patch)\n"
"end\n"
"\n"
"local function upgrade(options)\n"
"    options = options or {}\n"
"    setmetatable(options, {__index = {auto = false}})\n"
"\n"
"    local version = get_version()\n"
"    if version < mkversion(1, 7, 5) then\n"
"        log.warn('can upgrade from 1.7.5 only')\n"
"        return\n"
"    end\n"
"\n"
"    local handlers = {\n"
"        {version = mkversion(1, 7, 6), func = upgrade_to_1_7_6, auto = true},\n"
"        {version = mkversion(1, 7, 7), func = upgrade_to_1_7_7, auto = true},\n"
"        {version = mkversion(1, 10, 0), func = upgrade_to_1_10_0, auto = true},\n"
"        {version = mkversion(1, 10, 2), func = upgrade_to_1_10_2, auto = true},\n"
"        {version = mkversion(2, 1, 0), func = upgrade_to_2_1_0, auto = true},\n"
"        {version = mkversion(2, 1, 1), func = upgrade_to_2_1_1, auto = true},\n"
"        {version = mkversion(2, 1, 2), func = upgrade_to_2_1_2, auto = true},\n"
"        {version = mkversion(2, 1, 3), func = upgrade_to_2_1_3, auto = true},\n"
"        {version = mkversion(2, 2, 1), func = upgrade_to_2_2_1, auto = true},\n"
"        {version = mkversion(2, 3, 0), func = upgrade_to_2_3_0, auto = true},\n"
"        {version = mkversion(2, 3, 1), func = upgrade_to_2_3_1, auto = true},\n"
"    }\n"
"\n"
"    for _, handler in ipairs(handlers) do\n"
"        if version >= handler.version then\n"
"            goto continue\n"
"        end\n"
"        if options.auto and not handler.auto then\n"
"            log.warn(\"cannot auto upgrade schema version to %s, \" ..\n"
"                     \"please call box.schema.upgrade() manually\",\n"
"                     handler.version)\n"
"            return\n"
"        end\n"
"        handler.func()\n"
"        log.info(\"set schema version to %s\", handler.version)\n"
"        box.space._schema:replace({'version',\n"
"                                   handler.version.major,\n"
"                                   handler.version.minor,\n"
"                                   handler.version.patch})\n"
"        ::continue::\n"
"    end\n"
"end\n"
"\n"
"local function bootstrap()\n"
"    set_system_triggers(false)\n"
"\n"
"    -- erase current schema\n"
"    erase()\n"
"    -- insert initial schema\n"
"    initial_1_7_5()\n"
"    -- upgrade schema to the latest version\n"
"    upgrade()\n"
"\n"
"    set_system_triggers(true)\n"
"\n"
"    -- save new bootstrap.snap\n"
"    box.snapshot()\n"
"end\n"
"\n"
"box.schema.upgrade = upgrade;\n"
"box.internal.bootstrap = bootstrap;\n"
""
;
