const char init_lua[] =
"\n"
"-- init.lua -- internal file\n"
"\n"
"local ffi = require('ffi')\n"
"ffi.cdef[[\n"
"struct type_info;\n"
"struct method_info;\n"
"\n"
"enum ctype {\n"
"    CTYPE_VOID = 0,\n"
"    CTYPE_INT,\n"
"    CTYPE_CONST_CHAR_PTR\n"
"};\n"
"\n"
"struct type_info {\n"
"    const char *name;\n"
"    const struct type_info *parent;\n"
"    const struct method_info *methods;\n"
"};\n"
"\n"
"enum { METHOD_ARG_MAX = 8 };\n"
"\n"
"struct method_info {\n"
"    const struct type_info *owner;\n"
"    const char *name;\n"
"    enum ctype rtype;\n"
"    enum ctype atype[METHOD_ARG_MAX];\n"
"    int nargs;\n"
"    bool isconst;\n"
"\n"
"    union {\n"
"        /* Add extra space to get proper struct size in C */\n"
"        void *_spacer[2];\n"
"    };\n"
"};\n"
"\n"
"double\n"
"tarantool_uptime(void);\n"
"typedef int32_t pid_t;\n"
"pid_t getpid(void);\n"
"void\n"
"tarantool_exit(int);\n"
"]]\n"
"\n"
"local fio = require(\"fio\")\n"
"\n"
"local soext = (jit.os == \"OSX\" and \"dylib\" or \"so\")\n"
"\n"
"local ROCKS_LIB_PATH = '.rocks/lib/tarantool'\n"
"local ROCKS_LUA_PATH = '.rocks/share/tarantool'\n"
"local LIB_TEMPLATES = { '\?.'..soext }\n"
"local LUA_TEMPLATES = { '\?.lua', '\?/init.lua' }\n"
"local ROCKS_LIB_TEMPLATES = { ROCKS_LIB_PATH .. '/\?.'..soext }\n"
"local ROCKS_LUA_TEMPLATES = { ROCKS_LUA_PATH .. '/\?.lua', ROCKS_LUA_PATH .. '/\?/init.lua' }\n"
"\n"
"local package_searchroot\n"
"\n"
"local function searchroot()\n"
"    return package_searchroot or fio.cwd()\n"
"end\n"
"\n"
"local function setsearchroot(path)\n"
"    if not path then\n"
"        -- Here we need to get this function caller's sourcedir.\n"
"        path = debug.sourcedir(3)\n"
"    elseif path == box.NULL then\n"
"        path = nil\n"
"    else\n"
"        assert(type(path) == 'string', 'Search root must be a string')\n"
"    end\n"
"    package_searchroot = path and fio.abspath(path)\n"
"end\n"
"\n"
"dostring = function(s, ...)\n"
"    local chunk, message = loadstring(s)\n"
"    if chunk == nil then\n"
"        error(message, 2)\n"
"    end\n"
"    return chunk(...)\n"
"end\n"
"\n"
"local fiber = require(\"fiber\")\n"
"os.exit = function(code)\n"
"    code = (type(code) == 'number') and code or 0\n"
"    ffi.C.tarantool_exit(code)\n"
"    -- Make sure we yield even if the code after\n"
"    -- os.exit() never yields. After on_shutdown\n"
"    -- fiber completes, we will never wake up again.\n"
"    while true do fiber.yield() end\n"
"end\n"
"\n"
"local function uptime()\n"
"    return tonumber(ffi.C.tarantool_uptime());\n"
"end\n"
"\n"
"local function pid()\n"
"    return tonumber(ffi.C.getpid())\n"
"end\n"
"\n"
"local function mksymname(name)\n"
"    local mark = string.find(name, \"-\")\n"
"    if mark then name = string.sub(name, mark + 1) end\n"
"    return \"luaopen_\" .. string.gsub(name, \"%.\", \"_\")\n"
"end\n"
"\n"
"local function load_lib(file, name)\n"
"    return package.loadlib(file, mksymname(name))\n"
"end\n"
"\n"
"local function load_lua(file)\n"
"    return loadfile(file)\n"
"end\n"
"\n"
"local function traverse_path(path)\n"
"    path = fio.abspath(path)\n"
"    local paths = { path }\n"
"\n"
"    while path ~= '/' do\n"
"        path = fio.dirname(path)\n"
"        table.insert(paths, path)\n"
"    end\n"
"\n"
"    return paths\n"
"end\n"
"\n"
"-- Generate a search function, which performs searching through\n"
"-- templates setup in options.\n"
"--\n"
"-- @param path_fn function which returns a base path for the\n"
"--     resulting template\n"
"-- @param templates table with lua search templates\n"
"-- @param need_traverse bool flag which tells search function to\n"
"--     build multiple paths by expanding base path up to the\n"
"--     root ('/')\n"
"-- @return a searcher function which builds a path template and\n"
"--     calls package.searchpath\n"
"local function gen_search_func(path_fn, templates, need_traverse)\n"
"    assert(type(path_fn) == 'function', 'path_fn must be a function')\n"
"    assert(type(templates) == 'table', 'templates must be a table')\n"
"\n"
"    return function(name)\n"
"        local path = path_fn() or '.'\n"
"        local paths = need_traverse and traverse_path(path) or { path }\n"
"\n"
"        local searchpaths = {}\n"
"\n"
"        for _, path in ipairs(paths) do\n"
"            for _, template in pairs(templates) do\n"
"                table.insert(searchpaths, fio.pathjoin(path, template))\n"
"            end\n"
"        end\n"
"\n"
"        local searchpath = table.concat(searchpaths, ';')\n"
"\n"
"        return package.searchpath(name, searchpath)\n"
"    end\n"
"end\n"
"\n"
"-- Compose a loader function from options.\n"
"--\n"
"-- @param search_fn function will be used to search a file from\n"
"--     path template\n"
"-- @param load_fn function will be used to load a file, found by\n"
"--     search function\n"
"-- @return function a loader, which first search for the file and\n"
"--     then loads it\n"
"local function gen_loader_func(search_fn, load_fn)\n"
"    assert(type(search_fn) == 'function', 'search_fn must be defined')\n"
"    assert(type(load_fn) == 'function', 'load_fn must be defined')\n"
"\n"
"    return function(name)\n"
"        if not name then\n"
"            return \"empty name of module\"\n"
"        end\n"
"        local file, err = search_fn(name)\n"
"        if not file then\n"
"            return err\n"
"        end\n"
"        local loaded, err = load_fn(file, name)\n"
"        if err == nil then\n"
"            return loaded\n"
"        else\n"
"            return err\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local search_lua = gen_search_func(searchroot, LUA_TEMPLATES)\n"
"local search_lib = gen_search_func(searchroot, LIB_TEMPLATES)\n"
"local search_rocks_lua = gen_search_func(searchroot, ROCKS_LUA_TEMPLATES, true)\n"
"local search_rocks_lib = gen_search_func(searchroot, ROCKS_LIB_TEMPLATES, true)\n"
"\n"
"local search_funcs = {\n"
"    search_lua,\n"
"    search_lib,\n"
"    search_rocks_lua,\n"
"    search_rocks_lib,\n"
"    function(name) return package.searchpath(name, package.path) end,\n"
"    function(name) return package.searchpath(name, package.cpath) end,\n"
"}\n"
"\n"
"local function search(name)\n"
"    if not name then\n"
"        return \"empty name of module\"\n"
"    end\n"
"    for _, searcher in ipairs(search_funcs) do\n"
"        local file = searcher(name)\n"
"        if file ~= nil then\n"
"            return file\n"
"        end\n"
"    end\n"
"    return nil\n"
"end\n"
"\n"
"-- loader_preload 1\n"
"table.insert(package.loaders, 2, gen_loader_func(search_lua, load_lua))\n"
"table.insert(package.loaders, 3, gen_loader_func(search_lib, load_lib))\n"
"table.insert(package.loaders, 4, gen_loader_func(search_rocks_lua, load_lua))\n"
"table.insert(package.loaders, 5, gen_loader_func(search_rocks_lib, load_lib))\n"
"-- package.path   6\n"
"-- package.cpath  7\n"
"-- croot          8\n"
"\n"
"package.search = search\n"
"package.searchroot = searchroot\n"
"package.setsearchroot = setsearchroot\n"
"\n"
"return {\n"
"    uptime = uptime;\n"
"    pid = pid;\n"
"}\n"
""
;
